import json
import pandas as pd

from nerblackbox.modules.ner_training.data_preprocessing.tools.encodings_dataset import (
    EncodingsDataset,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.csv_reader import (
    CsvReader,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example import (
    InputExample,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_examples_to_tensors import (
    InputExamplesToTensors,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.utils import (
    InputExamples,
)
from nerblackbox.modules.ner_training.annotation_tags.annotation import Annotation
from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger
from nerblackbox.tests.utils import PseudoDefaultLogger
from nerblackbox.modules.utils.util_functions import get_dataset_path
from torch.utils.data import DataLoader, Sampler, RandomSampler, SequentialSampler
from pkg_resources import resource_filename
from copy import deepcopy
from os.path import join, isfile
from typing import List, Dict, Tuple, Optional, Any, Union

PHASES = ["train", "val", "test"]
SENTENCES_ROWS_UNPRETOKENIZED = List[Dict[str, Any]]


class DataPreprocessor:
    def __init__(
        self,
        tokenizer,
        do_lower_case: bool,
        default_logger: Union[DefaultLogger, PseudoDefaultLogger],
        max_seq_length: int = 64,
    ):
        """
        :param tokenizer:      [transformers Tokenizer]
        :param do_lower_case:  [bool] if True, make text data lowercase
        :param default_logger: [DefaultLogger]
        :param max_seq_length: [int], e.g. 64
        """
        self.tokenizer = tokenizer
        self.do_lower_case = do_lower_case
        self.default_logger = default_logger
        self.max_seq_length = max_seq_length
        self.pretokenized: Optional[
            bool
        ] = None  # whether dataset is pretokenized (csv) or not (jsonl)

    def get_input_examples_train(
        self,
        prune_ratio: Dict[str, float],
        dataset_name: Optional[str] = None,
        train_on_val: Optional[bool] = None,
        train_on_test: Optional[bool] = None,
    ) -> Tuple[Dict[str, InputExamples], Annotation]:
        """
        - get input examples for TRAIN from csv files

        Args:
            prune_ratio:      [dict], e.g. {'train': 1.0, 'val': 1.0, 'test': 1.0} -- pruning ratio for data
            dataset_name:     [str], e.g. 'suc'
            train_on_val:     [bool] if True, train additionally on validation set
            train_on_test:    [bool] if True, train additionally on test set

        Returns:
            input_examples:          [dict] w/ keys = 'train', 'val', 'test' & values = [list] of [InputExample]
            annotation:              [Annotation] instance
        """
        if dataset_name is None:
            dataset_path = resource_filename("nerblackbox", "tests/test_data")
        else:
            dataset_path = get_dataset_path(dataset_name)

        self.pretokenized = self._check_if_data_is_pretokenized(dataset_path)
        if not self.pretokenized:
            self._pretokenize(
                dataset_path
            )  # "{phase}.jsonl" -> "pretokenized_{phase}.csv"

        csv_reader = CsvReader(
            dataset_path,
            self.tokenizer,
            pretokenized=self.pretokenized,
            do_lower_case=self.do_lower_case,  # can be True (applies .lower()) !!
            default_logger=self.default_logger,
        )

        input_examples = dict()
        for phase in ["train", "val", "test"]:
            # train data
            input_examples_all = csv_reader.get_input_examples(phase)
            input_examples[phase] = self._prune_examples(
                input_examples_all, phase, ratio=prune_ratio[phase]
            )

        # note: w/o deepcopy, potential annotation scheme conversion fails
        # (as it will be applied twice in InputExamplesUtils.convert_annotation_scheme())
        if train_on_val:
            input_examples["train"] += deepcopy(input_examples["val"])
        if train_on_test:
            input_examples["train"] += deepcopy(input_examples["test"])

        annotation = Annotation(csv_reader.annotation_classes)

        return input_examples, annotation

    def get_input_examples_predict(
        self,
        examples: List[str],
        is_pretokenized: bool,
    ) -> Tuple[
        Dict[str, InputExamples], List[str], Optional[List[List[Tuple[int, int]]]]
    ]:
        """
        - get input examples for PREDICT from input argument examples

        Args:
            examples:       [list] of [str]
            is_pretokenized: True if examples are pretokenized

        Returns:
            input_examples:         [dict] w/ key = 'predict' & value = [list] of [InputExample]
            examples_pretokenized:  [list] of [str]
            pretokenization_offsets: [list] of [list] of [tuples]
        """
        # create input_examples
        if self.do_lower_case:
            examples = [example.lower() for example in examples]

        if is_pretokenized:
            _data_pretokenized = [
                {
                    "text": example,
                    "tags": " ".join(
                        ["O" for _ in range(len(example.split()))]
                    ),  # pseudo tags
                }
                for example in examples
            ]
            pretokenization_offsets = None
        else:
            _data = [
                {
                    "text": example,
                    "tags": [],
                }
                for example in examples
            ]
            _data_pretokenized, pretokenization_offsets = self._pretokenize_data(_data)

        examples_pretokenized = [elem["text"] for elem in _data_pretokenized]

        input_examples = {
            "predict": [
                InputExample(
                    guid="",
                    text=elem["text"],
                    tags=elem["tags"],
                )
                for elem in _data_pretokenized
            ]
        }

        return input_examples, examples_pretokenized, pretokenization_offsets

    def to_dataloader(
        self,
        input_examples: Dict[str, InputExamples],
        annotation_classes: List[str],
        batch_size: int,
    ) -> Tuple[Dict[str, DataLoader], Dict[str, List[int]]]:
        """
        - turn input_examples into dataloader

        Args:
            input_examples:     [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
                                          values = [list] of [InputExample]
            annotation_classes: [list] of tags present in the dataset, e.g. ['O', 'PER', ..]
            batch_size:         [int]

        Returns:
            _dataloader:    [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
                                      values = [torch Dataloader]
            _offsets:        [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
                                       values = [List] of [int]
                                       which gives information on how the input_examples where slices,
                                       e.g. offsets = [0, 2, 3]
                                       means that the first two slices belong to the first input example (0->2),
                                       while the third slice belongs to the second input example (2->3).
        """
        # input_example_to_tensors
        input_examples_to_tensors = InputExamplesToTensors(
            self.tokenizer,
            max_seq_length=self.max_seq_length,
            annotation_classes_tuple=tuple(annotation_classes),
            default_logger=self.default_logger,
        )

        _dataloader = dict()
        _offsets = dict()
        for phase in input_examples.keys():
            self.default_logger.log_info(
                f"[before preprocessing] {phase.ljust(5)} data: {len(input_examples[phase])} examples"
            )
            encodings, _offsets_phase = input_examples_to_tensors(
                input_examples[phase], predict=phase == "predict"
            )
            _offsets[phase] = _offsets_phase

            # dataloader
            data = EncodingsDataset(
                encodings=encodings
            )  # data[j] = 4 torch tensors corresponding to EncodingKeys
            if self.default_logger:
                self.default_logger.log_info(
                    f"[after  preprocessing] {phase.ljust(5)} data: {len(data)} examples"
                )

            assert phase in [
                "train",
                "val",
                "test",
                "predict",
            ], f"ERROR! phase = {phase} unknown."
            sampler: Optional[Sampler]
            if phase == "train":
                sampler = RandomSampler(data)
            elif phase in ["val", "test"]:
                sampler = SequentialSampler(data)
            else:  # if phase == "predict":
                sampler = None

            _dataloader[phase] = DataLoader(
                data,
                sampler=sampler,
                batch_size=batch_size,
                num_workers=0,  # num_workers=1, pin_memory=True
                # see https://pytorch-lightning.readthedocs.io/en/stable/benchmarking/performance.html
            )

        return _dataloader, _offsets

    ####################################################################################################################
    # HELPER
    ####################################################################################################################
    @staticmethod
    def _check_if_data_is_pretokenized(dataset_path: str) -> bool:
        """
        check if dataset_name is pretokenized (csv) or not (jsonl)

        Args:
            dataset_path: path to dataset directory

        Returns:
            pretokenized: whether dataset is pretokenized (csv) or not (jsonl)
        """
        dataset_path_csv = join(dataset_path, f"train.csv")
        dataset_path_jsonl = join(dataset_path, f"train.jsonl")
        if isfile(dataset_path_csv) and not isfile(dataset_path_jsonl):
            return True
        elif not isfile(dataset_path_csv) and isfile(dataset_path_jsonl):
            return False
        else:
            raise Exception(
                f"ERROR! Did not find train.csv OR train.jsonl in {dataset_path}:"
                f"(csv: {isfile(dataset_path_csv)}, jsonl: {isfile(dataset_path_jsonl)})"
            )

    @staticmethod
    def _tokens2words(_tokens: List[str]) -> List[str]:
        """
        merges tokens to words

        Args:
            _tokens: e.g. ["this", "example", "contains", "hu", "##gging", "face"]

        Returns:
            _words: e.g. ["this", "example", "contains", "hugging", "face"]
        """
        i = 1
        while i < len(_tokens):
            if _tokens[i].startswith("##"):
                _tokens[i - 1] += _tokens.pop(i).strip("##")  # join with previous
            else:
                i += 1
        return _tokens

    @staticmethod
    def _resolve_overlapping_tags(_tags: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """

        Args:
            _tags: e.g.
            [
             {"token": "Bajo peso", "tag": "Concept", "char_start": 4651, "char_end": 4660},
             {"token": "peso", "tag": "Concept", "char_start": 4656, "char_end": 4660},
             {"token": "más", "tag": "Predicate", "char_start": 4681, "char_end": 4684},
            ]

        Returns:
            _resolved_tags: e.g.
            [
             {"token": "Bajo peso", "tag": "Concept", "char_start": 4651, "char_end": 4660},
             {"token": "más", "tag": "Predicate", "char_start": 4681, "char_end": 4684},
            ]
        """
        _resolved_tags = list()
        for i in range(len(_tags)):
            if i == 0 or _tags[i - 1]["char_end"] <= _tags[i]["char_start"]:
                _resolved_tags.append(_tags[i])
        return _resolved_tags

    def _pretokenize_data(
        self, _data: SENTENCES_ROWS_UNPRETOKENIZED
    ) -> Tuple[List[Dict[str, str]], List[List[Tuple[int, int]]]]:
        """
        pretokenize data (text and text simultaneously)

        Args:
            _data: e.g. [
                {
                    'text': 'arbetsförmedlingen ai-center finns i stockholm.',
                    'tags': [
                        {"token": "arbetsförmedlingen ai-center", "tag": "ORG", "char_start": 0, "char_end": 28},
                        {"token": "stockholm", "tag": "LOC", "char_start": 37, "char_end": 46},
                    ]
                },
                ..
            ]

        Returns:
            _data_pretokenized: e.g. [
                {
                    'text': 'arbetsförmedlingen ai - center finns i stockholm .',
                    'tags': 'B-ORG I-ORG I-ORG I-ORG O O B-LOC O'
                },
                ..
            ]
            _pretokenization_offsets: e.g. [
                [(0,17), (18,20), (20,21), (21,26)]
            ]
        """
        _data_pretokenized = list()
        _pretokenization_offsets = list()
        for n in range(len(_data)):
            word_tuples = (
                self.tokenizer.backend_tokenizer.pre_tokenizer.pre_tokenize_str(
                    _data[n]["text"]
                )
            )
            tags = ["O"] * len(word_tuples)
            entity_dicts = self._resolve_overlapping_tags(_data[n]["tags"])
            for entity_dict in entity_dicts:
                entity_char_start = entity_dict["char_start"]
                entity_char_end = entity_dict["char_end"]
                entity_tag = entity_dict["tag"]
                for word_index, word_tuple in enumerate(word_tuples):
                    word_char_start = word_tuple[1][0]
                    word_char_end = word_tuple[1][1]
                    if (
                        word_char_start == entity_char_start
                        and word_char_end <= entity_char_end
                    ):
                        tags[word_index] = f"B-{entity_tag}"
                    elif (
                        word_char_start >= entity_char_start
                        and word_char_end <= entity_char_end
                    ):
                        tags[word_index] = f"I-{entity_tag}"

            words = [word_tuple[0] for word_tuple in word_tuples]
            pretokenization_offsets = [word_tuple[1] for word_tuple in word_tuples]
            _pretokenization_offsets.append(pretokenization_offsets)

            _data_pretokenized.append(
                {
                    "tags": " ".join(tags),
                    "text": " ".join(words),
                }
            )

        return _data_pretokenized, _pretokenization_offsets

    def _pretokenize(self, dataset_path: str) -> None:  # pragma: no cover
        """
        1. read jsonl files "{phase}.jsonl"
        2. pretokenize data
        3. write csv files "pretokenized_{phase}.csv"

        Args:
            dataset_path: path to dataset directory

        Returns: -
        """
        for phase in PHASES:
            # 1. read json files
            dataset_path_jsonl = join(dataset_path, f"{phase}.jsonl")
            with open(dataset_path_jsonl, "r") as f:
                jlines = f.readlines()
                data = [json.loads(jline) for jline in jlines]

            # 2. pretokenize data
            data_pretokenized, _ = self._pretokenize_data(data)

            # 3. write csv files "pretokenized_{phase}.csv"
            df = pd.DataFrame(data_pretokenized)
            file_path = join(dataset_path, f"pretokenized_{phase}.csv")
            df.to_csv(file_path, sep="\t", header=False, index=False)

    def _prune_examples(self, list_of_examples: List[Any], phase: str, ratio: float):
        """
        prunes list_of_examples by taking only the first examples
        ---------------------------------------------------------
        :param list_of_examples: [list], e.g. of [InputExample]
        :param phase:            [str], 'train' or 'valid'
        :param (Optional) ratio: [float], e.g. 0.5
        :return: [list], e.g. of [InputExample]
        """
        if ratio == 1.0:
            return list_of_examples
        else:
            num_examples_old = len(list_of_examples)
            num_examples_new = max(1, int(ratio * float(num_examples_old)))
            info = f"> {phase.ljust(5)} data: use {num_examples_new} of {num_examples_old} examples"
            if self.default_logger:
                self.default_logger.log_info(info)
            return list_of_examples[:num_examples_new]



import json
import string
from os.path import join, isdir, isfile
from typing import Tuple, Any, Optional
import numpy as np

import torch
from transformers import AutoModelForTokenClassification, AutoTokenizer
from typing import List, Union, Dict
from torch.nn.functional import softmax
from huggingface_hub import hf_hub_download
from itertools import product

from nerblackbox.modules.ner_training.annotation_tags.token_tags import TokenTags
from nerblackbox.modules.ner_training.data_preprocessing.data_preprocessor import (
    DataPreprocessor,
)
from nerblackbox.tests.utils import PseudoDefaultLogger
from nerblackbox.api.store import Store
from nerblackbox.api.dataset import Dataset
from nerblackbox.modules.ner_training.metrics.ner_metrics import NerMetrics
from nerblackbox.modules.experiment_results import ExperimentResults
from nerblackbox.modules.ner_training.ner_model_train2model import (
    NerModelTrain2Model,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.csv_reader import (
    CsvReader,
)
from nerblackbox.modules.ner_training.annotation_tags.tags import Tags

PREDICTIONS = List[List[Dict[str, Any]]]
EVALUATION_DICT = Dict[str, Dict[str, Dict[str, Optional[float]]]]

VERBOSE = False


class Model:
    r"""
    model that predicts tags for given input text
    """

    @classmethod
    def from_experiment(cls, experiment_name: str) -> Optional["Model"]:
        r"""load best model from experiment.

        Args:
            experiment_name: name of the experiment, e.g. "exp0"

        Returns:
            model: best model from experiment
        """
        experiment_exists, experiment_results = Store.get_experiment_results_single(
            experiment_name
        )
        assert (
            experiment_exists
        ), f"ERROR! experiment = {experiment_name} does not exist."

        assert isinstance(
            experiment_results, ExperimentResults
        ), f"ERROR! experiment_results expected to be an instance of ExperimentResults."

        if experiment_results.best_single_run is None:
            print(
                f"> ATTENTION! could not find results for experiment = {experiment_name}"
            )
            return None
        elif (
            "checkpoint" not in experiment_results.best_single_run.keys()
            or experiment_results.best_single_run["checkpoint"] is None
        ):
            print(
                f"> ATTENTION! there is no checkpoint for experiment = {experiment_name}."
            )
            return None
        else:
            checkpoint_path_train = experiment_results.best_single_run["checkpoint"]
            checkpoint_path_predict = checkpoint_path_train.strip(".ckpt")

            # translate NerModelTrain checkpoint to Model checkpoint if necessary
            if not Model.checkpoint_exists(checkpoint_path_predict):
                ner_model_train2model = NerModelTrain2Model.load_from_checkpoint(
                    checkpoint_path_train
                )
                ner_model_train2model.export_to_ner_model_prod(checkpoint_path_train)

            return Model(checkpoint_path_predict)

    @classmethod
    def from_checkpoint(cls, checkpoint_directory: str) -> Optional["Model"]:
        r"""

        Args:
            checkpoint_directory: path to the checkpoint directory

        Returns:
            model: best model from experiment
        """
        if not cls.checkpoint_exists(checkpoint_directory):
            print(
                f"> ATTENTION! could not find checkpoint directory at {checkpoint_directory}"
            )
            return None
        else:
            return Model(checkpoint_directory)

    @classmethod
    def from_huggingface(cls, repo_id: str) -> Optional["Model"]:
        r"""

        Args:
            repo_id: id of the huggingface hub repo id, e.g. 'KB/bert-base-swedish-cased-ner'

        Returns:
            model: model
        """
        # download files and get local file paths in cache directory
        filenames = [
            "config.json",
            "pytorch_model.bin",
            "special_tokens_map.json",
            "tokenizer_config.json",
            "vocab.txt",
        ]
        local_file_paths = []
        for filename in filenames:
            local_file_path = hf_hub_download(repo_id=repo_id, filename=filename)
            local_file_paths.append(local_file_path)
        assert len(local_file_paths) == len(
            filenames
        ), f"ERROR! #local_file_paths = {len(local_file_paths)} does not correspond to #files = {len(filenames)}."

        # extract cache directory
        cache_directories = [
            "/".join(local_file_path.split("/")[:-1])
            for local_file_path in local_file_paths
        ]
        assert (
            len(set(cache_directories)) == 1
        ), f"ERROR! cache directory could not be found due to inconsistency."
        cache_directory = cache_directories[0]

        # create Model from files in cache directory
        return Model(cache_directory)

    @classmethod
    def checkpoint_exists(cls, checkpoint_directory: str) -> bool:
        return isdir(checkpoint_directory)

    def __init__(
        self,
        checkpoint_directory: str,
        batch_size: int = 16,
        max_seq_length: Optional[int] = None,
    ):
        r"""
        Args:
            checkpoint_directory: path to the checkpoint directory
            batch_size: batch size used by dataloader
            max_seq_length: maximum sequence length (Optional). Loaded from checkpoint if not specified.
        """
        # 0. device
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        # 1: config (max_seq_length & annotation)
        path_config = join(checkpoint_directory, "config.json")
        with open(path_config, "r") as f:
            config = json.load(f)

        if max_seq_length:
            self.max_seq_length = max_seq_length
        elif "max_seq_length" in config.keys():
            self.max_seq_length = config["max_seq_length"]
        else:
            self.max_seq_length = config["max_position_embeddings"]

        id2label = {int(_id): label for _id, label in config["id2label"].items()}
        label2id = {label: int(_id) for _id, label in config["id2label"].items()}
        self.annotation_classes = list(config["id2label"].values())

        self.annotation_scheme = derive_annotation_scheme(id2label)

        # 3. model
        self.model = AutoModelForTokenClassification.from_pretrained(
            checkpoint_directory,
            id2label=id2label,
            label2id=label2id,
            return_dict=False,
        )
        self.model.eval()
        self.model = self.model.to(self.device)

        # 4. tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(
            checkpoint_directory,
        )

        # 5. batch_size (dataloader)
        self.batch_size = batch_size

    def predict_on_file(self, input_file: str, output_file: str) -> None:
        r"""
        predict tags for all input texts in input file, write results to output file

        Args:
            input_file: e.g. strangnas/test.jsonl
            output_file: e.g. strangnas/test_anonymized.jsonl
        """
        print(f"> read   input_file = {input_file}")
        with open(input_file, "r") as f:
            input_lines = [json.loads(line) for line in f]

        output_lines = list()
        for input_line in input_lines:
            text = input_line["text"]
            tags = self.predict(text)[0]
            output_line = {
                "text": text,
                "tags": tags,
            }
            output_lines.append(output_line)

        print(f"> write output_file = {output_file}")
        with open(output_file, "w") as f:
            for line in output_lines:
                f.write(json.dumps(line, ensure_ascii=False) + "\n")

    def predict(
        self,
        input_texts: Union[str, List[str]],
        level: str = "entity",
        autocorrect: bool = False,
        is_pretokenized: bool = False,
    ) -> PREDICTIONS:
        r"""predict tags for input texts. output on entity or word level.

        Examples:
            ```
            predict(["arbetsförmedlingen finns i stockholm"], level="word", autocorrect=False)
            # [[
            #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "tag": "I-ORG"},
            #     {"char_start": "19", "char_end": "24", "token": "finns", "tag": "O"},
            #     {"char_start": "25", "char_end": "26", "token": "i", "tag": "O"},
            #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "B-LOC"},
            # ]]
            ```
            ```
            predict(["arbetsförmedlingen finns i stockholm"], level="word", autocorrect=True)
            # [[
            #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "tag": "B-ORG"},
            #     {"char_start": "19", "char_end": "24", "token": "finns", "tag": "O"},
            #     {"char_start": "25", "char_end": "26", "token": "i", "tag": "O"},
            #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "B-LOC"},
            # ]]
            ```
            ```
            predict(["arbetsförmedlingen finns i stockholm"], level="entity", autocorrect=False)
            # [[
            #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "LOC"},
            # ]]
            ```
            ```
            predict(["arbetsförmedlingen finns i stockholm"], level="entity", autocorrect=True)
            # [[
            #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "tag": "ORG"},
            #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "LOC"},
            # ]]
            ```

        Args:
            input_texts:   e.g. ["example 1", "example 2"]
            level:         "entity" or "word"
            autocorrect:   if True, autocorrect annotation scheme (e.g. B- and I- tags).
            is_pretokenized: True if input_texts are pretokenized

        Returns:
            predictions: [list] of predictions for the different examples.
                         each list contains a [list] of [dict] w/ keys = char_start, char_end, word, tag
        """
        return self._predict(
            input_texts,
            level,
            autocorrect,
            proba=False,
            is_pretokenized=is_pretokenized,
        )

    def predict_proba(
        self, input_texts: Union[str, List[str]], is_pretokenized: bool = False
    ) -> PREDICTIONS:
        r"""predict probability distributions for input texts. output on word level.

        Examples:
            ```
            predict_proba(["arbetsförmedlingen finns i stockholm"])
            # [[
            #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "proba_dist: {"O": 0.21, "B-ORG": 0.56, ..}},
            #     {"char_start": "19", "char_end": "24", "token": "finns", "proba_dist: {"O": 0.87, "B-ORG": 0.02, ..}},
            #     {"char_start": "25", "char_end": "26", "token": "i", "proba_dist: {"O": 0.95, "B-ORG": 0.01, ..}},
            #     {"char_start": "27", "char_end": "36", "token": "stockholm", "proba_dist: {"O": 0.14, "B-ORG": 0.22, ..}},
            # ]]
            ```

        Args:
            input_texts:   e.g. ["example 1", "example 2"]
            is_pretokenized: True if input_texts are pretokenized

        Returns:
            predictions: [list] of probability predictions for different examples.
                         each list contains a [list] of [dict] w/ keys = char_start, char_end, word, proba_dist
                         where proba_dist = [dict] that maps self.annotation.classes to probabilities
        """
        return self._predict(
            input_texts,
            level="word",
            autocorrect=False,
            proba=True,
            is_pretokenized=is_pretokenized,
        )

    def _predict(
        self,
        input_texts: Union[str, List[str]],
        level: str = "entity",
        autocorrect: bool = False,
        proba: bool = False,
        is_pretokenized: bool = False,
    ) -> PREDICTIONS:
        r"""predict tags or probabilities for tags

        Args:
            input_texts:  e.g. ["example 1", "example 2"]
            level:        "entity" or "word"
            autocorrect:  if True, autocorrect annotation scheme (e.g. B- and I- tags).
            proba:        if True, predict probabilities instead of labels (on word level)
            is_pretokenized: True if input_texts are pretokenized

        Returns:
            predictions: [list] of [list] of [dict] w/ keys = char_start, char_end, word, tag/proba_dist
                         where proba_dist = [dict] that maps self.annotation.classes to probabilities
        """
        # --- check input arguments ---
        assert level in [
            "entity",
            "word",
        ], f"ERROR! model prediction level = {level} unknown, needs to be entity or word."
        if proba:
            assert level == "word" and autocorrect is False, (
                f"ERROR! probability predictions require level = word and autocorrect = False. "
                f"level = {level} and autocorrect = {autocorrect} not possible."
            )
        # ------------------------------

        # 0. ensure input_texts is a list
        if isinstance(input_texts, str):
            input_texts = [input_texts]
        number_of_input_texts = len(input_texts)

        # 1. pure model inference
        self.data_preprocessor = DataPreprocessor(
            tokenizer=self.tokenizer,
            do_lower_case=False,
            max_seq_length=self.max_seq_length,
            default_logger=PseudoDefaultLogger(),
        )
        (
            input_examples,
            input_texts_pretokenized,
            pretokenization_offsets,
        ) = self.data_preprocessor.get_input_examples_predict(
            examples=input_texts,
            is_pretokenized=is_pretokenized,
        )

        dataloader_all, offsets_all = self.data_preprocessor.to_dataloader(
            input_examples, self.annotation_classes, batch_size=self.batch_size
        )
        dataloader = dataloader_all["predict"]
        offsets = offsets_all["predict"]

        inputs_batches = list()
        outputs_batches = list()

        for sample in dataloader:
            inputs_batches.append(sample["input_ids"])
            sample.pop("labels")
            sample = {k: v.to(self.device) for k, v in sample.items()}
            with torch.no_grad():
                outputs_batch = self.model(**sample)[
                    0
                ]  # shape = [batch_size, seq_length, num_labels]
            outputs_batches.append(outputs_batch.detach().cpu())

        inputs = torch.cat(inputs_batches)  # tensor of shape = [batch_size, seq_length]
        outputs = torch.cat(
            outputs_batches
        )  # tensor of shape = [batch_size, seq_length, num_labels]

        ################################################################################################################
        ################################################################################################################
        tokens: List[List[str]] = [
            self.tokenizer.convert_ids_to_tokens(input_ids)
            for input_ids in inputs.tolist()
        ]  # List[List[str]] with len = batch_size

        predictions: List[List[Any]]
        if proba:
            predictions = turn_tensors_into_tag_probability_distributions(
                annotation_classes=self.annotation_classes,
                outputs=outputs,
            )  # List[List[Dict[str, float]]] with len = batch_size
        else:
            predictions_ids = torch.argmax(
                outputs, dim=2
            )  # tensor of shape [batch_size, seq_length]
            predictions = [
                [
                    self.model.config.id2label[prediction]
                    for prediction in predictions_ids[i].numpy()
                ]
                for i in range(len(predictions_ids))
            ]  # List[List[str]] with len = batch_size

        # merge
        tokens = [
            merge_slices_for_single_document(tokens[offsets[i] : offsets[i + 1]])
            for i in range(len(offsets) - 1)
        ]  # List[List[str]] with len = number_of_input_texts
        predictions = [
            merge_slices_for_single_document(predictions[offsets[i] : offsets[i + 1]])
            for i in range(len(offsets) - 1)
        ]  # List[List[str]] or List[List[Dict[str, float]]] with len = number_of_input_texts

        assert len(tokens) == len(
            predictions
        ), f"ERROR! len(tokens) = {len(tokens)} should equal len(predictions) = {len(predictions)}"
        assert (
            len(tokens) == number_of_input_texts
        ), f"ERROR! len(tokens) = {len(tokens)} should equal len(input_texts) = {number_of_input_texts}"
        assert (
            len(predictions) == number_of_input_texts
        ), f"ERROR! len(predictions) = {len(predictions)} should equal len(input_texts) = {number_of_input_texts}"

        # 2. post processing
        predictions = [
            self._post_processing(
                level,
                autocorrect,
                proba,
                input_texts[i],
                input_texts_pretokenized[i],
                pretokenization_offsets[i]
                if pretokenization_offsets is not None
                else None,
                tokens[i],
                predictions[i],
            )
            for i in range(number_of_input_texts)
        ]

        return predictions

    def _post_processing(
        self,
        level: str,
        autocorrect: bool,
        proba: bool,
        input_text: str,
        input_text_pretokenized: str,
        pretokenization_offsets: Optional[List[Tuple[int, int]]],
        tokens: List[str],
        predictions: List[Any],
    ) -> List[List[Dict[str, str]]]:
        r"""
        Args:
            level: "word" or "entity"
            autocorrect: e.g. False
            proba: e.g. False
            input_text: e.g. "we are in stockholm."
            input_text_pretokenized: e.g. "we are in stockholm ."
            pretokenization_offsets: e.g. [(0,2), (3,6), (7,9), (10,19), (19,20)]
            tokens: e.g. ["we", "are", "in", "stockholm", "."]
            predictions: e.g. ["O", "O", "O", "B-LOC", "O"]

        Returns:
            input_text_word_predictions: ???
        """
        ######################################
        # 1 input_text, merged chunks, tokens -> words
        ######################################
        _token_predictions: List[
            Tuple[str, Union[str, Dict[str, float]]]
        ] = merge_subtoken_to_token_predictions(tokens, predictions)

        token_predictions: List[Dict[str, Union[str, Dict]]] = restore_unknown_tokens(
            _token_predictions, input_text_pretokenized, verbose=VERBOSE
        )

        if autocorrect or level == "entity":
            assert (
                proba is False
            ), f"ERROR! autocorrect = {autocorrect} / level = {level} not allowed if proba = {proba}"

        #######################################
        token_predictions_str: List[Dict[str, str]] = assert_typing(token_predictions)
        token_tags = TokenTags(token_predictions_str, scheme=self.annotation_scheme)
        token_tags.merge_tokens_to_words()
        #######################################
        if pretokenization_offsets is not None:
            token_tags.unpretokenize(pretokenization_offsets)

        if autocorrect:
            token_tags.restore_annotation_scheme_consistency()

        if level == "entity":
            token_tags.merge_tokens_to_entities(
                original_text=input_text, verbose=VERBOSE
            )

        predictions = token_tags.as_list()

        return predictions

    def evaluate_on_dataset(
        self,
        dataset_name: str,
        dataset_format: str = "infer",
        phase: str = "test",
        class_mapping: Optional[Dict[str, str]] = None,
        number: Optional[int] = None,
        derived_from_jsonl: bool = False,
        rounded_decimals: Optional[int] = 3,
    ) -> EVALUATION_DICT:
        r"""
        evaluate model on dataset from huggingface or local dataset in jsonl or csv format

        Args:
            dataset_name: e.g. 'conll2003'
            dataset_format: 'huggingface', 'jsonl', 'csv'
            phase: e.g. 'test'
            class_mapping: e.g. {"PER": "PI", "ORG": "PI}
            number: e.g. 100
            derived_from_jsonl:
            rounded_decimals: if not None, results will be rounded to provided decimals

        Returns:
            evaluation_dict:
                Dict with keys [label][level][metric]
                where label in ['micro', 'macro'],
                level in ['entity', 'token']
                metric in ['precision', 'recall', 'f1', 'precision_seqeval', 'recall_seqeval', 'f1_seqeval']
                and values = float between 0 and 1

        """
        dataset_formats = ["infer", "jsonl", "csv", "huggingface"]
        phases = ["train", "val", "test"]
        assert (
            dataset_format in dataset_formats
        ), f"ERROR! dataset_format={dataset_format} unknown (known={dataset_formats})"
        assert phase in phases, f"ERROR! phase = {phase} unknown (known={phases})"

        store_path = Store.get_path()
        assert isinstance(
            store_path, str
        ), f"ERROR! type(store_path) = {type(store_path)} shoudl be str."

        if dataset_format == "infer":
            file_path_jsonl = join(
                store_path, "datasets", dataset_name, f"{phase}.jsonl"
            )
            file_path_csv = join(store_path, "datasets", dataset_name, f"{phase}.csv")
            if isfile(file_path_jsonl):
                dataset_format = "jsonl"
            elif isfile(file_path_csv):
                dataset_format = "csv"
            else:
                dataset_format = "huggingface"
            print(f"> dataset_format = {dataset_format} (inferred)")

        dir_path = join(store_path, "datasets", dataset_name)
        if dataset_format == "huggingface":
            evaluation_dict = self._evaluate_on_huggingface(
                dataset_name, phase, class_mapping, number
            )
        elif dataset_format == "jsonl":
            evaluation_dict = self._evaluate_on_jsonl(
                dir_path, phase, class_mapping, number
            )
        elif dataset_format == "csv":
            evaluation_dict = self._evaluate_on_csv(
                dir_path, phase, class_mapping, number, derived_from_jsonl
            )
        else:
            raise Exception(f"ERROR! dataset_format = {dataset_format} unknown.")

        if rounded_decimals is None:
            return evaluation_dict
        else:
            return round_evaluation_dict(evaluation_dict, rounded_decimals)

    def _evaluate_on_huggingface(
        self,
        dataset_name: str,
        phase: str,
        class_mapping: Optional[Dict[str, str]] = None,
        number: Optional[int] = None,
    ) -> EVALUATION_DICT:
        r"""
        evaluate model on dataset from huggingface

        Args:
            dataset_name: e.g. 'conll2003'
            phase: e.g. 'test'
            class_mapping: e.g. {"PER": "PI", "ORG": "PI}
            number: e.g. 100

        Returns:
            evaluation_dict:
                Dict with keys [label][level][metric]
                where label in ['micro', 'macro'],
                level in ['entity', 'token']
                metric in ['precision', 'recall', 'f1', 'precision_seqeval', 'recall_seqeval', 'f1_seqeval']
                and values = float between 0 and 1
        """
        dataset = Dataset(dataset_name)
        dataset.set_up()
        dir_path = f"{Store.get_path()}/datasets/{dataset_name}"
        return self._evaluate_on_csv(
            dir_path,
            phase=phase,
            class_mapping=class_mapping,
            number=number,
            derived_from_jsonl=False,
        )

    def _evaluate_on_jsonl(
        self,
        dir_path: str,
        phase: str,
        class_mapping: Optional[Dict[str, str]] = None,
        number: Optional[int] = None,
    ) -> EVALUATION_DICT:
        r"""
        evaluate model on local dataset in jsonl format

        Args:
            dir_path: e.g. './store/datasets/my_dataset'
            phase: e.g. 'test'
            class_mapping: e.g. {"PER": "PI", "ORG": "PI"}
            number: e.g. 100

        Returns:
            evaluation_dict:
                Dict with keys [label][level][metric]
                where label in ['micro', 'macro'],
                level in ['entity', 'token']
                metric in ['precision', 'recall', 'f1', 'precision_seqeval', 'recall_seqeval', 'f1_seqeval']
                and values = float between 0 and 1
        """
        assert isdir(dir_path), f"ERROR! could not find {dir_path}"
        data_preprocessor = DataPreprocessor(
            tokenizer=self.tokenizer,
            do_lower_case=False,
            max_seq_length=self.max_seq_length,
            default_logger=PseudoDefaultLogger(),
        )
        data_preprocessor._pretokenize(dir_path)

        return self._evaluate_on_csv(
            dir_path,
            phase=phase,
            class_mapping=class_mapping,
            number=number,
            derived_from_jsonl=True,
        )

    def _evaluate_on_csv(
        self,
        dir_path: str,
        phase: str,
        class_mapping: Optional[Dict[str, str]] = None,
        number: Optional[int] = None,
        derived_from_jsonl: bool = False,
    ) -> EVALUATION_DICT:
        r"""
        evaluate model on local dataset in csv format

        Args:
            dir_path: e.g. './store/datasets/my_dataset'
            phase: e.g. 'test'
            class_mapping: e.g. {"PER": "PI", "ORG": "PI}
            number: e.g. 100
            derived_from_jsonl: should be True is csv was created from jsonl through pretokenization

        Returns:
            evaluation_dict:
                Dict with keys [label][level][metric]
                where label in ['micro', 'macro'],
                level in ['entity', 'token']
                metric in ['precision', 'recall', 'f1', 'precision_seqeval', 'recall_seqeval', 'f1_seqeval']
                and values = float between 0 and 1
        """
        # derived_from_jsonl = True  => pretokenized_<phase>.csv is used
        # derived_from_jsonl = False => <phase>.csv is used
        file_path = join(
            dir_path,
            f"pretokenized_{phase}.csv" if derived_from_jsonl else f"{phase}.csv",
        )
        assert isfile(file_path), f"ERROR! could not find {file_path}"

        csv_reader = CsvReader(
            dir_path,
            self.tokenizer,
            pretokenized=derived_from_jsonl is False,
            do_lower_case=self.tokenizer.do_lower_case,
            default_logger=None,
        )
        data = csv_reader.get_input_examples(phase)
        ground_truth = [elem.tags.split() for elem in data]
        input_texts = [elem.text for elem in data]
        if number is not None:
            ground_truth = ground_truth[:number]
            input_texts = input_texts[:number]

        _predictions = self.predict(input_texts, level="word", is_pretokenized=True)
        predictions = [
            [elem["tag"] for elem in _prediction] for _prediction in _predictions
        ]

        def convert_plain_to_bio(_predictions: List[str]) -> List[str]:
            tags = Tags(_predictions)
            return tags.convert_scheme("plain", "bio")

        if self.annotation_scheme == "plain":
            print(
                "> ATTENTION! predictions converted from plain to bio annotation scheme!"
            )
            predictions = [
                convert_plain_to_bio(prediction) for prediction in predictions
            ]

        # check that ground truth and predictions have same lengths
        assert len(ground_truth) == len(
            predictions
        ), f"ERROR! #ground_truth = {len(ground_truth)}, #predictions = {len(predictions)}"
        for i in range(len(ground_truth)):
            assert len(ground_truth[i]) == len(predictions[i]), (
                f"ERROR! #ground_truth[{i}] = {len(ground_truth[i])} ({ground_truth[i]}), "
                f"#predictions[{i}] = {len(predictions[i])} ({predictions[i]}),"
                f"input_texts[{i}] = {input_texts[i]}"
            )

        return self._evaluate(ground_truth, predictions, class_mapping)

    @staticmethod
    def _evaluate(
        ground_truth: List[List[str]],
        predictions: List[List[str]],
        class_mapping: Optional[Dict[str, str]] = None,
    ) -> EVALUATION_DICT:
        r"""
        evaluate by comparing ground_truth with predictions (after applying class_mapping)

        Args:
            ground_truth: e.g. [["B-PER", "I-PER"], ["O", "B-ORG"]]
            predictions: e.g. [["B-PER", "O"], ["O", "B-ORG"]]
            class_mapping: e.g. {"PER": "PI", "ORG": "PI}

        Returns:
            evaluation_dict:
                Dict with keys [label][level][metric]
                where label in ['micro', 'macro'],
                level in ['entity', 'token']
                metric in ['precision', 'recall', 'f1', 'precision_seqeval', 'recall_seqeval', 'f1_seqeval']
                and values = float between 0 and 1
        """
        # class mapping
        def map_class(_class: str) -> str:
            r"""
            maps class according to class_mapping
            Args:
                _class: e.g. "B-PER"

            Returns:
                _class_new: e.g. "B-PI"

            """
            assert isinstance(
                class_mapping, dict
            ), f"ERROR! type(class_mapping) = {type(class_mapping)} should be dict"
            if _class == "O":
                return "O"
            else:
                assert (
                    "-" in _class
                ), f"ERROR! class = {_class} expected to be of BIO scheme."
                _class_prefix = _class.split("-")[0]
                _class_plain = _class.split("-")[-1]
                if _class_plain in class_mapping.keys():
                    _class_plain_new = class_mapping[_class_plain]
                    _class_new = f"{_class_prefix}-{_class_plain_new}"
                else:
                    _class_new = "O"
                return _class_new

        if class_mapping is not None:
            predictions = [
                [map_class(elem) for elem in sublist] for sublist in predictions
            ]

        # flatten
        true_flat = [elem for sublist in ground_truth for elem in sublist]
        pred_flat = [elem for sublist in predictions for elem in sublist]
        assert len(true_flat) == len(
            pred_flat
        ), f"ERROR! true_flat = {len(true_flat)}, #pred_flat = {len(pred_flat)}"

        # 4. evaluate: compare ground truth with predictions
        # NerMetrics
        labels = ["micro", "macro"]
        metrics = ["precision", "recall", "f1"]
        metrics_seqeval = [f"{metric}_seqeval" for metric in metrics]
        levels = ["entity", "token"]
        evaluation: EVALUATION_DICT = {
            label: {
                level: {metric: None for metric in metrics + metrics_seqeval}
                for level in levels
            }
            for label in labels
        }

        ner_metrics_entity = NerMetrics(
            true_flat, pred_flat, level="entity", scheme="bio"
        )
        ner_metrics_entity.compute(metrics)
        # print("== ENTITY (nerblackbox) ==")
        for metric in metrics:
            _metric = metric if metric == "acc" else f"{metric}_micro"
            # print(f"> {metric}: {ner_metrics_entity.results_as_dict()[_metric]:.3f}")
        # print()

        for metric, label in product(metrics, labels):
            evaluation[label]["entity"][metric] = ner_metrics_entity.results_as_dict()[
                f"{metric}_{label}"
            ]

        # seqeval - just for testing - start
        from seqeval.metrics import precision_score, recall_score, f1_score

        scores = [precision_score, recall_score, f1_score]
        # print("== ENTITY (seqeval) ==")
        for metric_seqeval, score in zip(metrics_seqeval, scores):
            result = score(
                ground_truth,
                predictions,
            )
            # print(f"> {metric_seqeval}: {result:.3f}")
            evaluation["micro"]["entity"][f"{metric_seqeval}"] = result

        return evaluation


########################################################################################################################
########################################################################################################################
########################################################################################################################
def round_evaluation_dict(
    _evaluation_dict: EVALUATION_DICT, _rounded_decimals: int
) -> EVALUATION_DICT:
    r"""

    Args:
        _evaluation_dict: e.g.
        {
            "micro": {
                "entity": {
                    'precision': 0.9847222222222223,
                    'recall': 0.9916083916083916,
                    'f1': 0.9881533101045297,
                    'precision_seqeval': 0.9833564493758669,
                    'recall_seqeval': 0.9916083916083916,
                    'f1_seqeval': 0.9874651810584958,
                }
            }
        }
        _rounded_decimals: e.g. 3

    Returns:
        _evaluation_dict_rounded: e.g.
        {
            "micro": {
                "entity": {
                    'precision': 0.985,
                    'recall': 0.992,
                    'f1': 0.988,
                    'precision_seqeval': 0.983,
                    'recall_seqeval': 0.992,
                    'f1_seqeval': 0.987,
                }
            }
        }
    """
    return {
        k1: {
            k2: {
                k3: round(v3, _rounded_decimals) if isinstance(v3, float) else None
                for k3, v3 in v2.items()
            }
            for k2, v2 in v1.items()
        }
        for k1, v1 in _evaluation_dict.items()
    }


def derive_annotation_scheme(_id2label: Dict[int, str]) -> str:
    r"""
    Args:
        _id2label: e.g. {0: "O", 1: "B-PER", 2: "I-PER"}

    Returns:
        annotation_scheme: e.g. "bio"
    """
    if len(_id2label) == 0:
        raise Exception(
            f"ERROR! could not derive annotation scheme. id2label is empty."
        )

    label_prefixes = list(
        set([label.split("-")[0] for label in _id2label.values() if "-" in label])
    )
    if len(label_prefixes) == 0:
        return "plain"
    elif (
        "B" in label_prefixes
        and "I" in label_prefixes
        and "L" in label_prefixes
        and "U" in label_prefixes
    ):
        return "bilou"
    elif "B" in label_prefixes and "I" in label_prefixes:
        return "bio"
    else:
        raise Exception(
            f"ERROR! could not derive annotation scheme. found label_prefixes = {label_prefixes}"
        )


def turn_tensors_into_tag_probability_distributions(
    annotation_classes: List[str], outputs: torch.Tensor
) -> List[List[Dict[str, float]]]:
    """
    Args:
        annotation_classes: e.g. ["O", "B-PER", "I-PER"]
        outputs: [torch tensor]  of shape = [batch_size, seq_length, num_labels]

    Returns:
        predictions_proba: [list] of [list] of [prob dist], i.e. dict that maps tags to probabilities
    """
    probability_distributions = softmax(outputs, dim=2)
    predictions_proba = [
        [
            {
                annotation_classes[j]: float(
                    probability_distributions[x][i][j].detach().numpy()
                )
                for j in range(len(annotation_classes))
            }
            for i in range(outputs.shape[1])  # seq_length
        ]
        for x in range(outputs.shape[0])  # number_of_input_texts
    ]

    return predictions_proba


def merge_slices_for_single_document(_list: List[List[Any]]) -> List[Any]:
    """
    merges the slices for a single document

    input is
    - either List[List[str]], for tokens & predictions
    - or     List[List[Dict[str, float]]], for predictions with proba = True

    Args:
        _list: predictions for one or more slices that belong to same document,
               e.g. one slice:  [["[CLS]", "this", "is", "one", "slice", "[SEP]"]],
               e.g. two slices: [
                                    ["[CLS]", "this", "is", "one", "slice", "[SEP]"],
                                    ["[CLS]", "and", "a", "second", "one", "[SEP]"]
                                ],

    Returns:
        _list_flat: predictions for one document
               e.g. one slice:  ["[CLS]", "this", "is", "one", "slice", "[SEP]"],
               e.g. two slices: ["[CLS]", "this", "is", "one", "slice", "and", "a", "second", "one", "[SEP]"],
    """
    if len(_list) == 1:  # one slice
        return _list[0]
    else:  # more slices
        _list_flat = list()
        for i, sublist in enumerate(_list):
            if i == 0:
                _list_flat.extend(sublist[:-1])
            elif 0 < i < len(_list) - 1:
                _list_flat.extend(sublist[1:-1])
            else:
                _list_flat.extend(sublist[1:])
        return _list_flat


def merge_subtoken_to_token_predictions(
    tokens: List[str],
    predictions: List[Any],
) -> List[Tuple[str, Any]]:
    """
    Args:
        tokens: e.g. ["[CLS]", "arbetsförmedl", "##ingen", "finns", "i", "stockholm", "[SEP]", "[PAD]"]
        predictions:  e.g. ["[S]", "ORG", "ORG", "O", "O", "O", "[S]", "[S]"]

    Returns:
        token_predictions: e.g. [("arbetsförmedlingen", "ORG"), ("finns", "O"), ("i", "O"), ("stockholm", "O")]
    """
    # predict tags on words between [CLS] and [SEP]
    token_predictions_list = list()
    for token, example_token_prediction in zip(tokens, predictions):
        if token not in ["[CLS]", "[SEP]", "[PAD]"]:
            if not token.startswith("##"):
                token_predictions_list.append([token, example_token_prediction])
            else:
                token_predictions_list[-1][0] += token.strip("##")

    token_predictions = [(elem[0], elem[1]) for elem in token_predictions_list]

    return token_predictions


def restore_unknown_tokens(
    word_predictions: List[Tuple[str, Union[str, Dict[str, float]]]],
    input_text: str,
    verbose: bool = False,
) -> List[Dict[str, Union[str, Dict]]]:
    """
    - replace "[UNK]" tokens by the original token
    - enrich tokens with char_start & char_end

    Args:
        word_predictions: e.g. [('example', 'O'), ('of', 'O), ('author', 'PERSON'), ..]
        input_text: e.g. 'example of author'
        verbose: e.g. False

    Returns:
        word_predictions_restored: e.g. [
            {"char_start": "0", "char_end": "7", "token": "example", "tag": "O"},
            ..
        ]
    """
    word_predictions_restored = list()

    # 1. get margins of known tokens
    token_char_margins: List[Tuple[Any, ...]] = list()
    char_start = 0
    unknown_counter = 0
    for token, _ in word_predictions:
        if token == "[UNK]":
            token_char_margins.append((None, None))
            unknown_counter += 1
        else:
            # move char_start to where punctuation or whitespace is found
            while unknown_counter > 0:
                try:
                    if token in string.punctuation or len(token) != 1:
                        char_start = input_text.index(token, char_start)
                    else:  # i.e. len(token) == 1
                        char_start = input_text.index(f" {token}", char_start)
                except ValueError:  # .index() did not find anything
                    if verbose:
                        print(
                            f"! token = {token} not found in example[{char_start}:] (unknown)"
                        )
                unknown_counter -= 1

            # find token_char_margins for token
            try:
                char_start = input_text.index(token, char_start)
                token_char_margins.append((char_start, char_start + len(token)))
            except ValueError:
                if verbose:
                    print(f"! token = {token} not found in example[{char_start}:]")
                token_char_margins.append((None, None))
            char_start += len(token)
            unknown_counter = 0

    # 2. restore unknown tokens
    for i, (token, tag) in enumerate(word_predictions):
        if (
            token_char_margins[i][0] is not None
            and token_char_margins[i][1] is not None
        ):
            word_predictions_restored.append(
                {
                    "char_start": str(token_char_margins[i][0]),
                    "char_end": str(token_char_margins[i][1]),
                    "token": token,
                    "tag": tag,
                }
            )
        else:
            for j in range(2):
                assert (
                    token_char_margins[i][j] is None
                ), f"ERROR! token_char_margin[{i}][{j}] is not None for token = {token}"

            char_start_margin, char_end_margin = None, None
            k_prev, k_next = None, None
            for k in range(1, 10):
                k_prev = k
                if i - k < 0:
                    char_start_margin = 0
                    break
                elif token_char_margins[i - k][1] is not None:
                    char_start_margin = token_char_margins[i - k][1]
                    break
            for k in range(1, 10):
                k_next = k
                if i + k >= len(token_char_margins):
                    char_end_margin = len(input_text)
                    break
                elif token_char_margins[i + k][0] is not None:
                    char_end_margin = token_char_margins[i + k][0]
                    break
            assert (
                char_start_margin is not None
            ), f"ERROR! could not find char_start_margin"
            assert char_end_margin is not None, (
                f"ERROR! could not find char_end_margin. token = {token}\n"
                f"char_start_margin = {char_start_margin}\n"
                f"word_predictions = {word_predictions}\n"
                f"input_text = {input_text}\n"
                f"token_char_margins = {token_char_margins}\n"
                f"word_predictions_restored = {word_predictions_restored}"
            )
            assert (
                k_prev is not None and k_next is not None
            ), f"ERROR! could not find k_prev or k_next"

            new_token = input_text[char_start_margin:char_end_margin].strip()
            if k_prev != 1 or k_next != 1:
                new_token_parts = new_token.split()
                assert len(new_token_parts) == 1 + np.absolute(k_prev - k_next), (
                    f"ERROR! new_token = {new_token} consists of {len(new_token_parts)} parts, "
                    f"expected {1+np.absolute(k_prev-k_next)} (k_prev = {k_prev}, k_next = {k_next})"
                )
                new_token = new_token_parts[k_prev - 1]

            if len(new_token):
                char_start = input_text.index(new_token, char_start_margin)
                char_end = char_start + len(new_token)
                if verbose:
                    print(
                        f"! restored unknown token = {new_token} between "
                        f"char_start = {char_start}, "
                        f"char_end = {char_end}"
                    )
                word_predictions_restored.append(
                    {
                        "char_start": str(char_start),
                        "char_end": str(char_end),
                        "token": new_token,
                        "tag": tag,
                    }
                )
            else:
                if verbose:
                    print(
                        f"! dropped unknown empty token between "
                        f"char_start = {char_start_margin}, "
                        f"char_end = {char_end_margin}"
                    )

    return word_predictions_restored


def assert_typing(
    input_text_word_predictions: List[Dict[str, Any]]
) -> List[Dict[str, str]]:
    """
    this is only to ensure correct typing, it does not actually change anything

    Args:
        input_text_word_predictions: e.g. [
            {"char_start": 0, "char_end": 7, "token": "example", "tag": "O"},
            ..
        ]

    Returns:
        input_text_word_predictions_str: e.g. [
            {"char_start": "0", "char_end": "7", "token": "example", "tag": "O"},
            ..
        ]
    """
    return [
        {k: str(v) for k, v in input_text_word_prediction.items()}
        for input_text_word_prediction in input_text_word_predictions
    ]



import torch
import mlflow
import os
from typing import Tuple, Dict, Any
from os.path import join
from pytorch_lightning import Trainer
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.callbacks.base import Callback
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from pytorch_lightning.utilities.seed import seed_everything

from nerblackbox.modules.ner_training.ner_model_train import (
    NerModelTrain,
)
from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger
from nerblackbox.modules.utils.util_functions import unify_parameters
from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.utils.util_functions import (
    get_package_version,
    checkpoint2epoch,
    epoch2checkpoint,
)
from nerblackbox.modules.ner_training.callbacks.custom_early_stopping import (
    CustomEarlyStopping,
)
from nerblackbox.modules.ner_training.callbacks.learning_rate_changer import (
    LearningRateChanger,
)


def execute_single_run(params, hparams, log_dirs, experiment: bool):
    """
    :param params:     [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param hparams:    [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, lr_*
    :param log_dirs:   [argparse.Namespace] attr: mlflow, tensorboard
    :param experiment: [bool] whether run is part of an experiment w/ multiple runs
    :return: -
    """
    # seed
    seed = params.seed + int(
        params.run_name_nr.split("-")[1]
    )  # run_name_nr = e.g. 'runA-1', 'runA-2', 'runB-1', ..
    seed_everything(seed)

    default_logger = DefaultLogger(
        __file__, log_file=log_dirs.log_file, level=params.logging_level
    )  # python logging
    default_logger.clear()

    print_run_information(params, hparams, default_logger, seed)

    lightning_hparams = unify_parameters(params, hparams, log_dirs, experiment)

    tb_logger = logging_start(params, log_dirs)
    with mlflow.start_run(run_name=params.run_name_nr, nested=experiment):

        model = NerModelTrain(lightning_hparams)
        callbacks = get_callbacks(params, hparams, log_dirs)

        trainer = Trainer(
            max_epochs=hparams.max_epochs,
            gpus=torch.cuda.device_count() if params.device.type == "cuda" else None,
            precision=16 if (params.fp16 and params.device.type == "cuda") else 32,
            # amp_level="O1",
            logger=tb_logger,
            callbacks=list(callbacks),
        )
        trainer.fit(model)
        callback_info = get_callback_info(callbacks, params, hparams)

        default_logger.log_info(
            f"---\n"
            f"LOAD BEST CHECKPOINT {callback_info['checkpoint_best']}\n"
            f"FOR TESTING AND DETAILED RESULTS\n"
            f"---"
        )
        model_best = NerModelTrain.load_from_checkpoint(
            checkpoint_path=callback_info["checkpoint_best"]
        )
        trainer_best = Trainer(
            gpus=torch.cuda.device_count() if params.device.type == "cuda" else None,
            precision=16 if (params.fp16 and params.device.type == "cuda") else 32,
            logger=tb_logger,
        )
        trainer_best.validate(model_best)
        trainer_best.test(model_best)

        # logging end
        logging_end(
            tb_logger, callback_info, hparams, model, model_best, default_logger
        )

        # remove checkpoint
        if params.checkpoints is False:
            remove_checkpoint(callback_info["checkpoint_best"], default_logger)


########################################################################################################################
# HELPER FUNCTIONS #####################################################################################################
########################################################################################################################
def print_run_information(_params, _hparams, _logger, _seed: int):
    """
    :param _params:   [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param _hparams:  [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, prune_ratio_*, lr_*
    :param _logger:   [DefaultLogger]
    :param _seed:     [int]
    :return: -
    """
    _logger.log_info(f">>> NER BLACK BOX VERSION: {get_package_version()}")
    _logger.log_info("- PARAMS -----------------------------------------")
    _logger.log_info(f"> experiment_name: {_params.experiment_name}")
    _logger.log_info(f"> from_config:     {_params.from_config}")
    _logger.log_info(f"> run_name_nr:     {_params.run_name_nr}")
    _logger.log_info("..")
    _logger.log_info(f"> available GPUs: {torch.cuda.device_count()}")
    _logger.log_info(f"> device:         {_params.device}")
    _logger.log_info(f"> fp16:           {_params.fp16}")
    _logger.log_info("..")
    _logger.log_info(f"> dataset_name:          {_params.dataset_name}")
    _logger.log_info(f"> annotation_scheme:     {_params.annotation_scheme}")
    _logger.log_info(f"> prune_ratio_train:     {_params.prune_ratio_train}")
    _logger.log_info(f"> prune_ratio_val:       {_params.prune_ratio_val}")
    _logger.log_info(f"> prune_ratio_test:      {_params.prune_ratio_test}")
    _logger.log_info(f"> train_on_val:          {_params.train_on_val}")
    _logger.log_info(f"> train_on_test:         {_params.train_on_test}")
    _logger.log_info("..")
    _logger.log_info(f"> pretrained_model_name: {_params.pretrained_model_name}")
    _logger.log_info(f"> uncased:               {_params.uncased}")
    _logger.log_info("..")
    _logger.log_info(f"> checkpoints:           {_params.checkpoints}")
    _logger.log_info(f"> logging_level:         {_params.logging_level}")
    _logger.log_info(f"> multiple_runs:         {_params.multiple_runs}")
    _logger.log_info(f"> seed:                  {_seed}")
    _logger.log_info("")
    _logger.log_info("- HPARAMS ----------------------------------------")
    _logger.log_info(f"> batch_size:           {_hparams.batch_size}")
    _logger.log_info(f"> max_seq_length:       {_hparams.max_seq_length}")
    _logger.log_info(f"> max_epochs:           {_hparams.max_epochs}")
    _logger.log_info(f"> early_stopping:       {_hparams.early_stopping}")
    _logger.log_info(f"> monitor:              {_hparams.monitor}")
    _logger.log_info(f"> min_delta:            {_hparams.min_delta}")
    _logger.log_info(f"> patience:             {_hparams.patience}")
    _logger.log_info(f"> mode:                 {_hparams.mode}")
    _logger.log_info(f"> lr_max:               {_hparams.lr_max}")
    _logger.log_info(f"> lr_warmup_epochs:     {_hparams.lr_warmup_epochs}")
    _logger.log_info(f"> lr_cooldown_epochs:   {_hparams.lr_cooldown_epochs}")
    _logger.log_info(f"> lr_cooldown_restarts: {_hparams.lr_cooldown_restarts}")
    _logger.log_info(f"> lr_schedule:          {_hparams.lr_schedule}")
    _logger.log_info(f"> lr_num_cycles:        {_hparams.lr_num_cycles}")
    _logger.log_info("")


def _get_model_checkpoint_directory(_params):
    """
    :param _params:     [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :return: model_checkpoint_directory [str]
    """
    return join(env_variable("DIR_CHECKPOINTS"), _params.experiment_run_name_nr)


def get_callbacks(_params, _hparams, _log_dirs) -> Tuple[Callback, ...]:
    """
    :param _params:     [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param _hparams:    [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, lr_*
    :param _log_dirs:   [argparse.Namespace] attr: mlflow, tensorboard
    :return: _callbacks: [list] w/ [pytorch lightning callback]
    """
    early_stopping = vars(_hparams)["early_stopping"]
    lr_schedule_hybrid = vars(_hparams)["lr_schedule"] == "hybrid"
    assert (
        early_stopping is False or lr_schedule_hybrid is False
    ), f"ERROR! early_stopping and lr_schedule_hybrid cannot both be True."

    model_checkpoint = ModelCheckpoint(
        dirpath=_get_model_checkpoint_directory(_params),
        filename="{epoch}",
        monitor=vars(_hparams)["monitor"],
        verbose=True,
        save_last=True,
    )
    model_checkpoint.CHECKPOINT_NAME_LAST = "{epoch}"

    _callbacks: Tuple[Callback, ...]
    if early_stopping:
        early_stopping_params = {
            k: vars(_hparams)[k]
            for k in [
                "monitor",
                "min_delta",
                "patience",
                "mode",
                "lr_schedule",
                "lr_cooldown_epochs",
                "lr_cooldown_restarts",
            ]
        }
        # e.g. patience = 0, lr_cooldown_epochs = 3 -> patience = 4
        early_stopping_params["patience"] += vars(_hparams)["lr_cooldown_epochs"] + 1

        _callbacks = (
            model_checkpoint,
            CustomEarlyStopping(
                **early_stopping_params,
                verbose=True,
            ),
        )
    elif lr_schedule_hybrid:
        learning_rate_changer_params = {
            k: vars(_hparams)[k]
            for k in [
                "max_epochs",
                "lr_cooldown_epochs",
            ]
        }
        _callbacks = (
            model_checkpoint,
            LearningRateChanger(
                **learning_rate_changer_params,
            ),
        )
    else:
        _callbacks = (model_checkpoint,)

    return _callbacks


def get_callback_info(_callbacks, _params, _hparams):
    """
    :param _callbacks: [Tuple], either (ModelCheckpoint) or (ModelCheckpoint, EarlyStopping)
    :param _params:    [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param _hparams:   [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, lr_*
    :return: _callback_info: [dict] w/ keys 'epoch_best', 'epoch_stopped', 'checkpoint_best'
    """
    callback_info: Dict[str, Any] = dict()

    checkpoint_best: str = _callbacks[0].last_model_path
    callback_info["epoch_best"] = checkpoint2epoch(checkpoint_best)

    callback_info["checkpoint_best"] = join(
        _get_model_checkpoint_directory(_params),
        epoch2checkpoint(callback_info["epoch_best"]),
    )

    early_stopping = len(_callbacks) > 1 and hasattr(_callbacks[1], "stopped_epoch")
    if early_stopping:
        callback_info["epoch_stopped"] = (
            callback_info["epoch_best"] - _hparams.lr_cooldown_epochs
        )
    else:
        callback_info["epoch_stopped"] = _hparams.max_epochs - 1

    return callback_info


def logging_start(_params, _log_dirs):
    """
    :param _params:      [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param _log_dirs:    [argparse.Namespace] attr: mlflow, tensorboard
    :return: _tb_logger: [pytorch lightning TensorBoardLogger]
    """
    # mlflow.tracking.set_tracking_uri(_log_dirs.mlflow)                # mlflow
    # mlflow.set_experiment(_params.experiment_name)                    # mlflow
    _tb_logger = TensorBoardLogger(
        save_dir=_log_dirs.tensorboard,  # tensorboard
        name=_params.experiment_run_name_nr,
    )
    return _tb_logger


def logging_end(
    _tb_logger, _callback_info, _hparams, _model_stopped, _model_best, _logger
):
    """
    :param _tb_logger:     [pytorch lightning TensorBoardLogger]
    :param _callback_info: [dict] w/ keys 'epoch_best', 'epoch_stopped', 'checkpoint_best'
    :param _hparams:       [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, lr_*
    :param _model_stopped: [NerModelTrain] epoch_stopped
    :param _model_best:    [NerModelTrain] epoch_best
    :param _logger:        [DefaultLogger]
    :return: -
    """
    epoch_best = _callback_info["epoch_best"]
    epoch_stopped = _callback_info["epoch_stopped"]

    _logger.log_info(f"epoch_best: {epoch_best}")
    _logger.log_info(f"epoch_stopped: {epoch_stopped}")

    _model_stopped.mlflow_client.log_metric("EPOCH_BEST", epoch_best)
    _model_stopped.mlflow_client.log_metric("EPOCH_STOPPED", epoch_stopped)

    for metric in ("token_fil_f1_micro", "entity_fil_f1_micro"):
        _model_stopped.mlflow_client.log_metric(
            f"epoch_best_val_{metric}".upper(),
            _model_best.epoch_metrics["val"][0][metric],
        )
        _model_stopped.mlflow_client.log_metric(
            f"epoch_best_test_{metric}".upper(),
            _model_best.epoch_metrics["test"][0][metric],
        )

    _model_stopped.mlflow_client.finish_artifact_logger()  # mlflow
    _tb_logger_stopped_epoch(
        _tb_logger, _hparams, epoch_best, epoch_stopped, _model_stopped, _model_best
    )  # tb


def _tb_logger_stopped_epoch(
    _tb_logger,
    _hparams,
    _epoch_best,
    _epoch_stopped,
    _model_stopped,
    _model_best,
    metrics=("token_fil_f1_micro", "entity_fil_f1_micro"),
):
    """
    log hparams and metrics for stopped epoch
    -----------------------------------------
    :param _tb_logger:      [pytorch lightning TensorboardLogger]
    :param _hparams:        [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, prune_ratio_*, lr_*
    :param _epoch_best:     [int]
    :param _epoch_stopped:  [int]
    :param _model_stopped:  [NerModelTrain] epoch_stopped
    :param _model_best:     [NerModelTrain] epoch_best
    :param metrics:         [tuple] of metrics to be logged in hparams section
    :return:
    """
    hparams_dict = dict()

    # epoch_best & epoch_stopped
    hparams_dict["hparam/train/epoch_best"] = _epoch_best
    hparams_dict["hparam/train/epoch_stopped"] = _epoch_stopped

    # val/test
    hparams_best_val = {
        f"hparam/val/epoch_best_{_epoch_best}/{metric}": _model_best.epoch_metrics[
            "val"
        ][0][metric]
        for metric in metrics
    }

    hparams_best_test = {
        f"hparam/test/epoch_best_{_epoch_best}/{metric}": _model_best.epoch_metrics[
            "test"
        ][0][metric]
        for metric in metrics
    }

    hparams_dict.update(hparams_best_val)
    hparams_dict.update(hparams_best_test)

    """
    NOTE: 
    The following command (= that adds hyperparameters (_hparams) & most important metrics (hparams_dict) to tb logging)
    logs properly to tensorboard's SCALAR, but not to HPARAMS.
    Apparently it is incompatible with pytorch-lightning's default tensorboard logging, 
    see https://github.com/PyTorchLightning/pytorch-lightning/issues/1225
    So technically speaking there are 2 logs for each run, pytorch-lightning's default & the one below. 
    """
    # log
    _tb_logger.experiment.add_hparams(
        vars(_hparams),
        hparams_dict,
    )


def remove_checkpoint(_checkpoint_path, _default_logger):
    """
    remove checkpoint stored at _checkpoint_path
    --------------------------------------------
    :param _checkpoint_path: [str], e.g. '[..]/store/results/checkpoints/exp_default/run1/epoch=0.ckpt'
    :param _default_logger:  [DefaultLogger]
    :return: -
    """
    os.remove(_checkpoint_path)
    _default_logger.log_info(f"> checkpoint {_checkpoint_path} removed")



import pytest
from typing import Optional


def pytest_approx(number: Optional[float]):
    """
    get acceptable pytest range for number
    --------------------------------------
    :param number: [float], e.g. 0.82
    :return: pytest range, e.g. 0.82 +- 0.01
    """
    if number is None:
        return None
    else:
        test_precision = 0.01
        return pytest.approx(number, abs=test_precision)



import pandas as pd
from typing import List, Dict, Tuple, Any, Union
from mlflow.entities import Run
from nerblackbox.modules.experiment_results import ExperimentResults

import os
from os.path import join, abspath, dirname, isfile
import pickle
from pkg_resources import resource_filename

BASE_DIR = abspath(dirname(dirname(dirname(__file__))))
DATA_DIR = join(BASE_DIR, "data")
os.environ["DATA_DIR"] = DATA_DIR
TRACKING_URI = resource_filename("nerblackbox", f"data/results/mlruns")
FILE_PATH = join(
    resource_filename("nerblackbox", f"tests/test_data"), "test_mlflow_runs.p"
)


########################################################################################################################
########################################################################################################################
########################################################################################################################
class TestExperimentResults:

    # load data
    if isfile(FILE_PATH):
        runs: List[Run] = pickle.load(open(FILE_PATH, "rb"))
    else:
        runs = None  # NO TESTING

    experiment_id: str = "0"
    experiment_name: str = "my_experiment_mlflow_runs"
    experiment_results = ExperimentResults(_id=experiment_id, name=experiment_name)

    # results: intermediate
    true_parameters_experiment: Dict[str, Any] = {
        "early_stopping": ["True"],
        "prune_ratio_test": ["0.2"],
        "seed": ["42"],
        "monitor": ["val_loss"],
        "lr_num_cycles": ["4"],
        "prune_ratio_val": ["0.2"],
        "pretrained_model_name": ["af-ai-center/bert-base-swedish-uncased"],
        "max_epochs": ["3"],
        "mode": ["min"],
        "patience": ["5"],
        "multiple_runs": ["2"],
        "checkpoints": ["True"],
        "lr_warmup_epochs": ["2"],
        "lr_cooldown": ["True"],
        "prune_ratio_train": ["0.2"],
        "annotation_scheme": ["plain"],
        "experiment_name": ["my_experiment_mlflow_runs"],
        "device": ["gpu"],
        "min_delta": ["0.3"],
        "dataset_name": ["swedish_ner_corpus"],
        "run_name": ["''"],
        "uncased": ["True"],
        "fp16": ["0"],
        "logging_level": ["info"],
    }
    true_parameters_runs: Dict[Tuple[str, str], Any] = {
        ("info", "run_id"): ["<run_id_2>", "<run_id_1>"],
        ("info", "run_name_nr"): ["runA-2", "runA-1"],
        ("params", "max_seq_length"): ["64", "64"],
        ("params", "lr_max"): ["2e-05", "2e-05"],
        ("params", "lr_schedule"): ["constant", "constant"],
        ("params", "batch_size"): ["16", "16"],
        ("metrics", "EPOCHS"): [3, 3],
        ("metrics", "EPOCH_BEST"): [2, 2],
        ("metrics", "EPOCH_STOPPED"): [2, 2],
        ("metrics", "EPOCH_BEST_VAL_ENTITY_FIL_F1_MICRO"): [
            0.7276507276507277,
            0.7757575757575756,
        ],
        ("metrics", "EPOCH_BEST_TEST_ENTITY_FIL_F1_MICRO"): [
            0.7388535031847134,
            0.691358024691358,
        ],
        ("metrics", "EPOCH_BEST_VAL_TOKEN_FIL_F1_MICRO"): [
            0.7879699248120301,
            0.8240469208211143,
        ],
        ("metrics", "EPOCH_BEST_TEST_TOKEN_FIL_F1_MICRO"): [
            0.7771260997067448,
            0.7463976945244957,
        ],
        ("metrics", "entity_fil_precision_micro"): [
            0.7981651376146789,
            0.721030042918455,
        ],
        ("metrics", "entity_fil_recall_micro"): [
            0.6877470355731226,
            0.6640316205533597,
        ],
        ("metrics", "entity_fil_asr_abidance"): [1.0, 1.0],
        ("metrics", "entity_fil_asr_precision_micro"): [
            0.7981651376146789,
            0.721030042918455,
        ],
        ("metrics", "entity_fil_asr_recall_micro"): [
            0.6877470355731226,
            0.6640316205533597,
        ],
        ("metrics", "entity_fil_asr_f1_micro"): [0.7388535031847134, 0.691358024691358],
    }
    true_parameters_runs_renamed: Dict[Tuple[str, str], Any] = {
        ("info", "run_id"): ["<run_id_2>", "<run_id_1>"],
        ("info", "run_name_nr"): ["runA-2", "runA-1"],
        ("params", "max_seq"): ["64", "64"],
        ("params", "lr_max"): ["2e-05", "2e-05"],
        ("params", "lr_sch"): ["constant", "constant"],
        ("params", "batch_sz"): ["16", "16"],
        ("metrics", "EPOCHS"): [3, 3],
        ("metrics", "EPOCH_BEST"): [2, 2],
        ("metrics", "EPOCH_STOPPED"): [2, 2],
        ("metrics", "VAL_ENT_F1"): [0.7276507276507277, 0.7757575757575756],
        ("metrics", "TEST_ENT_F1"): [0.7388535031847134, 0.691358024691358],
        ("metrics", "VAL_TOK_F1"): [0.7879699248120301, 0.8240469208211143],
        ("metrics", "TEST_TOK_F1"): [0.7771260997067448, 0.7463976945244957],
        ("metrics", "TEST_ENT_PRE"): [0.7981651376146789, 0.721030042918455],
        ("metrics", "TEST_ENT_REC"): [0.6877470355731226, 0.6640316205533597],
        ("metrics", "TEST_ENT_ASR_ABI"): [1.0, 1.0],
        ("metrics", "TEST_ENT_ASR_PRE"): [0.7981651376146789, 0.721030042918455],
        ("metrics", "TEST_ENT_ASR_REC"): [0.6877470355731226, 0.6640316205533597],
        ("metrics", "TEST_ENT_ASR_F1"): [0.7388535031847134, 0.691358024691358],
    }
    true_parameters_runs_renamed_average: Dict[Tuple[str, str], List[str]] = {
        ("info", "run_name"): ["runA"],
        ("params", "max_seq"): ["64"],
        ("params", "lr_max"): ["2e-05"],
        ("params", "lr_sch"): ["constant"],
        ("params", "batch_sz"): ["16"],
        ("metrics", "CONVERGENCE"): ["2/2"],
        ("metrics", "EPOCHS"): ["3.00000 +- 0.00000"],
        ("metrics", "EPOCH_BEST"): ["2.00000 +- 0.00000"],
        ("metrics", "EPOCH_STOPPED"): ["2.00000 +- 0.00000"],
        ("metrics", "VAL_ENT_F1"): ["0.75170 +- 0.01701"],
        ("metrics", "TEST_ENT_F1"): ["0.71511 +- 0.01679"],
        ("metrics", "VAL_TOK_F1"): ["0.80601 +- 0.01276"],
        ("metrics", "TEST_TOK_F1"): ["0.76176 +- 0.01086"],
        ("metrics", "TEST_ENT_PRE"): ["0.75960 +- 0.02727"],
        ("metrics", "TEST_ENT_REC"): ["0.67589 +- 0.00838"],
        ("metrics", "TEST_ENT_ASR_ABI"): ["1.00000 +- 0.00000"],
        ("metrics", "TEST_ENT_ASR_PRE"): ["0.75960 +- 0.02727"],
        ("metrics", "TEST_ENT_ASR_REC"): ["0.67589 +- 0.00838"],
        ("metrics", "TEST_ENT_ASR_F1"): ["0.71511 +- 0.01679"],
    }

    # results: ExperimentResults attributes
    true_experiment: pd.DataFrame = pd.DataFrame(
        true_parameters_experiment, index=["experiment"]
    ).T
    true_single_runs: pd.DataFrame = pd.DataFrame(
        {
            ("info", "run_id"): {1: "<run_id_2>", 0: "<run_id_1>"},
            ("info", "run_name_nr"): {1: "runA-1", 0: "runA-2"},
            ("params", "max_seq"): {1: "64", 0: "64"},
            ("params", "lr_max"): {1: "2e-05", 0: "2e-05"},
            ("params", "lr_sch"): {1: "constant", 0: "constant"},
            ("params", "batch_sz"): {1: "16", 0: "16"},
            ("metrics", "EPOCHS"): {1: 3, 0: 3},
            ("metrics", "EPOCH_BEST"): {1: 2, 0: 2},
            ("metrics", "EPOCH_STOPPED"): {1: 2, 0: 2},
            ("metrics", "VAL_ENT_F1"): {1: 0.7757575757575756, 0: 0.7276507276507277},
            ("metrics", "TEST_ENT_F1"): {1: 0.691358024691358, 0: 0.7388535031847134},
            ("metrics", "VAL_TOK_F1"): {1: 0.8240469208211143, 0: 0.7879699248120301},
            ("metrics", "TEST_TOK_F1"): {1: 0.7463976945244957, 0: 0.7771260997067448},
            ("metrics", "TEST_ENT_PRE"): {1: 0.721030042918455, 0: 0.7981651376146789},
            ("metrics", "TEST_ENT_REC"): {1: 0.6640316205533597, 0: 0.6877470355731226},
            ("metrics", "TEST_ENT_ASR_ABI"): {1: 1.0, 0: 1.0},
            ("metrics", "TEST_ENT_ASR_PRE"): {
                1: 0.721030042918455,
                0: 0.7981651376146789,
            },
            ("metrics", "TEST_ENT_ASR_REC"): {
                1: 0.6640316205533597,
                0: 0.6877470355731226,
            },
            ("metrics", "TEST_ENT_ASR_F1"): {
                1: 0.691358024691358,
                0: 0.7388535031847134,
            },
        }
    )
    true_average_runs: pd.DataFrame = pd.DataFrame(
        {
            ("info", "run_name"): {0: "runA"},
            ("params", "max_seq"): {0: "64"},
            ("params", "lr_max"): {0: "2e-05"},
            ("params", "lr_sch"): {0: "constant"},
            ("params", "batch_sz"): {0: "16"},
            ("metrics", "CONVERGENCE"): {0: "2/2"},
            ("metrics", "EPOCHS"): {0: "3.00000 +- 0.00000"},
            ("metrics", "EPOCH_BEST"): {0: "2.00000 +- 0.00000"},
            ("metrics", "EPOCH_BEST"): {0: "2.00000 +- 0.00000"},
            ("metrics", "EPOCH_STOPPED"): {0: "2.00000 +- 0.00000"},
            ("metrics", "VAL_ENT_F1"): {0: "0.75170 +- 0.01701"},
            ("metrics", "TEST_ENT_F1"): {0: "0.71511 +- 0.01679"},
            ("metrics", "VAL_TOK_F1"): {0: "0.80601 +- 0.01276"},
            ("metrics", "TEST_TOK_F1"): {0: "0.76176 +- 0.01086"},
            ("metrics", "TEST_ENT_PRE"): {0: "0.75960 +- 0.02727"},
            ("metrics", "TEST_ENT_REC"): {0: "0.67589 +- 0.00838"},
            ("metrics", "TEST_ENT_ASR_ABI"): {0: "1.00000 +- 0.00000"},
            ("metrics", "TEST_ENT_ASR_PRE"): {0: "0.75960 +- 0.02727"},
            ("metrics", "TEST_ENT_ASR_REC"): {0: "0.67589 +- 0.00838"},
            ("metrics", "TEST_ENT_ASR_F1"): {0: "0.71511 +- 0.01679"},
        }
    )
    true_best_single_run: Dict[str, Union[str, float]] = {
        "exp_id": "0",
        "exp_name": "my_experiment_mlflow_runs",
        "run_id": "<run_id_2>",
        "run_name_nr": "runA-1",
        "VAL_ENT_F1": 0.7757575757575756,
        "TEST_ENT_F1": 0.691358024691358,
        "VAL_TOK_F1": 0.8240469208211143,
        "TEST_TOK_F1": 0.7463976945244957,
        "TEST_ENT_PRE": 0.721030042918455,
        "TEST_ENT_REC": 0.6640316205533597,
        "TEST_ENT_ASR_ABI": 1.0,
        "TEST_ENT_ASR_PRE": 0.721030042918455,
        "TEST_ENT_ASR_REC": 0.6640316205533597,
        "TEST_ENT_ASR_F1": 0.691358024691358,
        "checkpoint": "<checkpoint>",
    }
    true_best_average_run: Dict[str, str] = {
        "exp_id": "0",
        "exp_name": "my_experiment_mlflow_runs",
        "run_name": "runA",
        "CONVERGENCE": "2/2",
        "VAL_ENT_F1": "0.75170 +- 0.01701",
        "TEST_ENT_F1": "0.71511 +- 0.01679",
        "VAL_TOK_F1": "0.80601 +- 0.01276",
        "TEST_TOK_F1": "0.76176 +- 0.01086",
        "TEST_ENT_PRE": "0.75960 +- 0.02727",
        "TEST_ENT_REC": "0.67589 +- 0.00838",
        "TEST_ENT_ASR_ABI": "1.00000 +- 0.00000",
        "TEST_ENT_ASR_PRE": "0.75960 +- 0.02727",
        "TEST_ENT_ASR_REC": "0.67589 +- 0.00838",
        "TEST_ENT_ASR_F1": "0.71511 +- 0.01679",
    }

    ####################################################################################################################
    # O. FROM_MLFLOW_RUNS
    ####################################################################################################################
    def test_from_mlflow_runs(self):
        if self.runs is not None:
            test_experiment_results: ExperimentResults = (
                self.experiment_results.from_mlflow_runs(
                    self.runs, self.experiment_id, self.experiment_name
                )
            )
            test_experiment_results.single_runs[("info", "run_id")] = [
                "<run_id_2>",
                "<run_id_1>",
            ]
            test_experiment_results.best_single_run["run_id"] = "<run_id_2>"
            test_experiment_results.best_single_run["checkpoint"] = "<checkpoint>"

            pd.testing.assert_frame_equal(
                test_experiment_results.experiment, self.true_experiment
            ), f"ERROR! test_parse_and_create_dataframe / experiment did not pass test"
            pd.testing.assert_frame_equal(
                test_experiment_results.single_runs, self.true_single_runs
            ), f"ERROR! test_parse_and_create_dataframe / single_runs did not pass test"
            pd.testing.assert_frame_equal(
                test_experiment_results.average_runs, self.true_average_runs
            ), f"ERROR! test_parse_and_create_dataframe / average_runs did not pass test"

            for k in self.true_best_single_run.keys():
                assert (
                    test_experiment_results.best_single_run[k]
                    == self.true_best_single_run[k]
                ), f"ERROR! test_best_single_run[{k}] = {test_experiment_results.best_single_run[k]} != {self.true_best_single_run[k]}"
            for k in self.true_best_average_run.keys():
                assert (
                    test_experiment_results.best_average_run[k]
                    == self.true_best_average_run[k]
                ), f"ERROR! test_best_average_run[{k}] = {test_experiment_results.best_average_run[k]} != {self.true_best_average_run[k]}"

    ####################################################################################################################
    # 1. PARSE AND CREATE DATAFRAME -> experiment, single_runs, average_runs
    ####################################################################################################################
    def test_parse_runs(self):
        if self.runs is not None:
            (
                test_parameters_runs,
                test_parameters_experiment,
            ) = self.experiment_results._parse_runs(self.runs)
            test_parameters_runs[("info", "run_id")] = ["<run_id_2>", "<run_id_1>"]

            for k1 in self.true_parameters_runs.keys():
                assert (
                    test_parameters_runs[k1] == self.true_parameters_runs[k1]
                ), f"ERROR! test_parameter_runs[{k1}] = {test_parameters_runs[k1]} != {self.true_parameters_runs[k1]}"

            for k2 in self.true_parameters_experiment.keys():
                assert (
                    test_parameters_experiment[k2]
                    == self.true_parameters_experiment[k2]
                ), f"ERROR! test_parameter_experiment[{k2}] = {test_parameters_experiment[k2]} != {self.true_parameters_experiment[k2]}"

    def test_rename_parameters_runs(self):
        if self.runs is not None:
            test_parameters_runs_renamed = (
                self.experiment_results._rename_parameters_runs(
                    self.true_parameters_runs
                )
            )

            for k in self.true_parameters_runs_renamed.keys():
                assert (
                    test_parameters_runs_renamed[k]
                    == self.true_parameters_runs_renamed[k]
                ), f"ERROR! test_parameter_runs_renamed[{k}] = {test_parameters_runs_renamed[k]} != {self.true_parameters_runs_renamed[k]}"

    def test_average(self):
        if self.runs is not None:
            test_parameters_runs_renamed_average = self.experiment_results._average(
                self.true_parameters_runs_renamed
            )
            for k in self.true_parameters_runs_renamed_average.keys():
                assert (
                    test_parameters_runs_renamed_average[k]
                    == self.true_parameters_runs_renamed_average[k]
                ), f"ERROR! test_parameter_runs_renamed_average[{k}] = {test_parameters_runs_renamed_average[k]} != {self.true_parameters_runs_renamed_average[k]}"

    def test_parse_and_create_dataframe(self):
        if self.runs is not None:
            self.experiment_results._parse_and_create_dataframe(
                self.runs
            )  # attr: experiment, single_runs, average_runs
            self.experiment_results.single_runs[("info", "run_id")] = [
                "<run_id_2>",
                "<run_id_1>",
            ]
            pd.testing.assert_frame_equal(
                self.experiment_results.experiment, self.true_experiment
            ), f"ERROR! test_parse_and_create_dataframe / experiment did not pass test"
            pd.testing.assert_frame_equal(
                self.experiment_results.single_runs, self.true_single_runs
            ), f"ERROR! test_parse_and_create_dataframe / single_runs did not pass test"
            pd.testing.assert_frame_equal(
                self.experiment_results.average_runs, self.true_average_runs
            ), f"ERROR! test_parse_and_create_dataframe / average_runs did not pass test"

    ####################################################################################################################
    # 2a. EXTRACT BEST SINGLE RUN -> best_single_run
    ####################################################################################################################
    def test_extract_best_single_run(self):
        if self.runs is not None:
            self.experiment_results.extract_best_single_run()  # attr: best_single_run
            self.experiment_results.best_single_run["checkpoint"] = "<checkpoint>"
            for k in self.true_best_single_run.keys():
                assert (
                    self.experiment_results.best_single_run[k]
                    == self.true_best_single_run[k]
                ), f"ERROR! test_best_single_run[{k}] = {self.experiment_results.best_single_run[k]} != {self.true_best_single_run[k]}"

    ####################################################################################################################
    # 2b. EXTRACT BEST AVERAGE RUN -> best_average_run
    ####################################################################################################################
    def test_extract_best_average_run(self):
        if self.runs is not None:
            self.experiment_results.extract_best_average_run()  # attr: best_average_run
            print("test_best_average_run")
            print(self.experiment_results.best_average_run)
            for k in self.true_best_average_run.keys():
                assert (
                    self.experiment_results.best_average_run[k]
                    == self.true_best_average_run[k]
                ), f"ERROR! test_best_average_run[{k}] = {self.experiment_results.best_average_run[k]} != {self.true_best_average_run[k]}"



import json
import os
from os.path import join, isdir, isfile

from transformers import AutoModelForTokenClassification
from omegaconf import DictConfig
from omegaconf.errors import ConfigKeyError
from transformers import AutoTokenizer
from nerblackbox.modules.ner_training.ner_model import NerModel

# TODO: this suppresses warnings (https://github.com/huggingface/transformers/issues/5421), root problem should be fixed
from transformers import logging

logging.set_verbosity_error()


class NerModelTrain2Model(NerModel):
    """
    class that translates (pytorch-lightning) NerModelTrain checkpoint to (transformers) NerModelPredict checkpoint
    """

    def __init__(self, hparams: DictConfig):
        """
        :param hparams: attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
        """
        super().__init__(hparams)

    def _preparations(self):
        assert isinstance(
            self.hparams, DictConfig
        ), f"ERROR! type(self.hparams) = {type(self.hparams)} should be a DictConfig."

        self.annotation_classes = json.loads(self.hparams["annotation_classes"])
        self.pretrained_model_name = self.hparams["pretrained_model_name"]
        try:
            self.encoding = json.loads(self.hparams["encoding"])
            self.special_tokens = list(set(self.encoding.values()))
        except ConfigKeyError:  # TODO: get rid of this exception handling.
            print(
                "> Note: special tokens are loaded instead of encoding (model was trained with nerblackbox<=0.0.12)"
            )
            self.encoding = None
            self.special_tokens = json.loads(self.hparams["special_tokens"])
        self.max_seq_length = int(self.hparams["max_seq_length"])
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.pretrained_model_name,
            do_lower_case=False,
            additional_special_tokens=self.special_tokens,
            use_fast=True,
        )  # do_lower_case needs to be False !!
        self.model = AutoModelForTokenClassification.from_pretrained(
            self.pretrained_model_name,
            num_labels=len(self.annotation_classes),
            return_dict=False,
        )
        self.model.resize_token_embeddings(
            len(self.tokenizer)
        )  # due to additional_special_tokens

    ####################################################################################################################
    # EXPORT TO NER_MODEL_PROD
    ####################################################################################################################
    def export_to_ner_model_prod(self, checkpoint_path: str):
        assert checkpoint_path.endswith(".p") or checkpoint_path.endswith(
            ".ckpt"
        ), f"ERROR! checkpoint_path = {checkpoint_path} is no *.p or *.ckpt file."
        export_directory = checkpoint_path.strip(".p").strip(".ckpt")
        assert not isfile(
            export_directory
        ), f"ERROR! export_directory = {export_directory} seems to be a file instead of a directory"

        if not isdir(export_directory):
            os.makedirs(export_directory, exist_ok=False)

            # 1. tokenizer
            self.tokenizer.save_pretrained(export_directory)

            # 2. model
            self.model.save_pretrained(export_directory)

            # 3. config (max_seq_length & annotation)
            path_config = join(export_directory, "config.json")
            assert isfile(
                path_config
            ), f"ERROR! config file at {path_config} does not exist."
            with open(path_config, "r") as f:
                config = json.load(f)

            config["max_seq_length"] = self.max_seq_length
            config["id2label"] = {
                _id: _label for _id, _label in enumerate(self.annotation_classes)
            }
            config["label2id"] = {
                _label: _id for _id, _label in enumerate(self.annotation_classes)
            }

            with open(path_config, "w") as f:
                json.dump(config, f)



import os
from os.path import isfile
import json
from typing import Optional
from nerblackbox.modules.datasets.formatter.auto_formatter import AutoFormatter

PHASES = ["train", "val", "test"]


class Dataset:
    r"""
    class to download, set up and inspect a single dataset
    """
    path = os.environ.get("DATA_DIR")

    def __init__(
        self,
        dataset_name: str,
        dataset_subset_name: str = "",
    ):
        r"""
        Args:
             dataset_name: name of dataset, e.g. "swedish_ner_corpus"
             dataset_subset_name: name of subset if applicable (for HuggingFace Datasets), e.g. "simple_cased"
        """
        self.dataset_name = dataset_name
        self.dataset_subset_name = dataset_subset_name
        self.file_path: Optional[
            str
        ] = None  # only for instances created through from_file method

    @classmethod
    def from_file(cls, dataset_name: str, file_path: str) -> Optional["Dataset"]:
        r"""
        create a Dataset instance by reading a file

        Args:
            dataset_name: name of dataset, e.g. "strangnas"
            file_path: e.g. "strangnas/strangnas.jsonl"

        Returns:
            dataset instance

        """
        if isfile(file_path):
            print(f"> create Dataset")
            dataset = Dataset(dataset_name)
            dataset.file_path = file_path
            return dataset
        else:
            print(f"Error! {file_path} does not exist")
            return None

    def split(self, val_fraction: float = 0.1, test_fraction: float = 0.2) -> None:
        r"""
        - read dataset from self.file_path
        - split dataset into train/val/test subsets
        - write subsets to PHASE.jsonl

        Args:
            val_fraction: e.g. 0.1
            test_fraction: e.g. 0.2
        """
        assert isinstance(
            self.file_path, str
        ), f"ERROR! type(self.file_path) = {self.file_path} should be str."

        with open(self.file_path, "r") as f:
            input_lines = [json.loads(line) for line in f]
        print(f"> read {len(input_lines)} documents from {self.file_path}")

        index_train_end = int((1 - val_fraction - test_fraction) * len(input_lines))
        index_val_end = int((1 - test_fraction) * len(input_lines))
        output_lines = {
            "train": input_lines[:index_train_end],
            "val": input_lines[index_train_end:index_val_end],
            "test": input_lines[index_val_end:],
        }

        print()
        for phase in PHASES:
            file_path_phase = (
                "/".join(self.file_path.split("/")[:-1]) + f"/{phase}.jsonl"
            )

            with open(file_path_phase, "w") as f:
                for line in output_lines[phase]:
                    f.write(json.dumps(line, ensure_ascii=False) + "\n")

            print(f"> wrote {len(output_lines[phase])} documents to {file_path_phase}")

    def set_up(
        self,
        modify: bool = True,
        val_fraction: float = 0.3,
        verbose: bool = False,
        shuffle: bool = False,
    ):
        r"""downloads and sets up the dataset. creates the following files:

            <STORE_DIR>/datasets/<dataset_name>/train.*
            <STORE_DIR>/datasets/<dataset_name>/val.*
            <STORE_DIR>/datasets/<dataset_name>/test.*

        where `* = json` or `* = csv`, depending on whether the data is pretokenized or not

        Args:
            modify: if True: modify tags as specified in method modify_ner_tag_mapping()  TODO: better explanation
            val_fraction: fraction of the validation dataset if it is split off from the training dataset
            verbose: verbose output
            shuffle: whether to shuffle train/val/test datasets
        """
        _parameters = {
            "ner_dataset": self.dataset_name,
            "ner_dataset_subset": self.dataset_subset_name,
            "modify": modify,
            "val_fraction": val_fraction,
            "verbose": verbose,
        }

        formatter = AutoFormatter.for_dataset(
            self.dataset_name, self.dataset_subset_name
        )
        formatter.create_directory()
        formatter.get_data(verbose=verbose)  # I: get_data
        formatter.create_ner_tag_mapping_json(
            modify=modify
        )  # II: create ner tag mapping
        formatter.format_data(shuffle=shuffle)  # III: format data
        formatter.resplit_data(
            val_fraction=val_fraction, write_csv=True
        )  # IV: resplit data
        formatter.analyzer.analyze_data()  # V: analyze data
        formatter.analyzer.plot_data()  # V: analyze data

    @staticmethod
    def overview():
        print("Dataset.overview() not yet implemented.")

    def _analyze(self):
        r"""
        analyze the dataset
        """
        formatter = AutoFormatter.for_dataset(self.dataset_name)
        formatter.analyzer.analyze_data()  # V: analyze data
        formatter.analyzer.plot_data()  # V: analyze data



from os.path import join, isfile
from typing import List, Dict, Optional, Tuple
import pandas as pd

from nerblackbox.modules.datasets.plots import Plots
from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger


class Analyzer:

    phases = ["train", "val", "test"]
    phases_all = ["total"] + phases

    def __init__(self, ner_dataset: str, ner_tag_list: List[str], dataset_path: str):
        """
        Args:
            ner_dataset:  'swedish_ner_corpus' or 'suc'
            ner_tag_list: e.g. ['PER', 'LOC', ..]
            dataset_path: e.g. DATA_DIR/datasets/swedish_ner_corpus
        """
        self.ner_dataset = ner_dataset
        self.ner_tag_list = ner_tag_list
        self.dataset_path = dataset_path
        self.stats_aggregated: Dict[str, Optional[pd.DataFrame]] = {"total": None}
        self.num_tokens: Dict[str, int] = {"total": 0}
        self.num_sentences: Dict[str, int] = {"total": 0}
        self.analysis_flag: bool = True

    def analyze_data(self, write_log: bool = True) -> None:
        """
        V: analyze data

        Args:
            write_log: whether to write log file (should always be True except for testing)

        Created Attr:
            stats_aggregated: [dict] w/ keys = 'total', 'train', 'val', 'test' & values = [df]
            num_tokens:       [dict] w/ keys = 'total', 'train', 'val', 'test' & values = [int]
            num_sentences:    [dict] w/ keys = 'total', 'train', 'val', 'test' & values = [int]

        Returns: -
        """
        for phase in self.phases:
            file_path = join(self.dataset_path, f"{phase}.csv")
            if not isfile(file_path):
                file_path_jsonl = join(self.dataset_path, f"{phase}.jsonl")
                if isfile(file_path_jsonl):
                    self.analysis_flag = False
                    return None
                else:
                    raise Exception(f"ERROR! file = {file_path} not found!")

            (
                self.num_sentences[phase],  # int, e.g. 4
                _stats_aggregated_phase,  # pd.DataFrame w/ col "tags", index = tags, values = tag occurrences
            ) = self._read_final_csv(phase)

            self.num_sentences["total"] += self.num_sentences[phase]
            if self.stats_aggregated["total"] is None:
                self.stats_aggregated["total"] = _stats_aggregated_phase
            else:
                self.stats_aggregated["total"] = (
                    self.stats_aggregated["total"] + _stats_aggregated_phase
                )

            self.stats_aggregated[phase] = self._stats_aggregated_extend(
                _stats_aggregated_phase, self.num_sentences[phase]
            )

        num_sentences_total = self.num_sentences["total"]
        self.stats_aggregated["total"] = self._stats_aggregated_extend(
            self.stats_aggregated["total"], self.num_sentences["total"]
        )

        # num_tokens
        self.num_tokens = {
            phase: self._get_num_tokens(self.stats_aggregated[phase])
            for phase in self.phases_all
        }
        num_tokens_total = self.num_tokens["total"]

        if write_log:  # pragma: no cover
            self._write_log_file(num_sentences_total, num_tokens_total)

    def plot_data(self) -> None:  # pragma: no cover
        if self.analysis_flag:
            fig_path = join(
                self.dataset_path, "analyze_data", f"{self.ner_dataset}.png"
            )
            Plots(self.stats_aggregated, self.num_sentences).plot(fig_path=fig_path)

    ####################################################################################################################
    # HELPER METHODS: analyze_data #####################################################################################
    ####################################################################################################################
    def _read_final_csv(self, phase: str) -> Tuple[int, pd.DataFrame]:
        """
        V: read final csv files

        Args:
            phase:         [str] 'train' or 'test'

        Returns:
            num_sentences:    [int]
            stats_aggregated: [pandas DataFrame] with indices = tags, column = "tag", values = number of tag occurrences
        """
        file_path = join(self.dataset_path, f"{phase}.csv")
        assert isfile(file_path), f"ERROR! file = {file_path} not found!"

        columns = ["O"] + self.ner_tag_list

        df = pd.read_csv(file_path, sep="\t", header=None, names=["labels", "text"])

        stats = pd.DataFrame([], columns=columns)

        if df is not None:
            tags = (
                df.iloc[:, 0]
                .apply(lambda x: str(x).split())
                .apply(lambda x: [elem.split("-")[-1] for elem in x])
            )

            for column in columns:
                stats[column] = tags.apply(
                    lambda x: len([elem for elem in x if elem == column])
                )

            assert len(df) == len(
                stats
            ), f"ERROR! len(df) = {len(df)} is not equal to len(stats) = {len(stats)}"

        num_sentences = len(stats)
        stats_aggregated = stats.sum().to_frame().astype(int)
        stats_aggregated.columns = ["tags"]

        return num_sentences, stats_aggregated

    @staticmethod
    def _stats_aggregated_extend(
        df: pd.DataFrame, number_of_sentences: int
    ) -> pd.DataFrame:
        """
        V: analyze data
        extend dataframe by adding columns

        Args:
            df: with column "tags"
            number_of_sentences: e.g. 1000

        Returns:
            df_extended: with columns ["tags", "tags/sentence", "tags relative w/ 0", "tags relative w/o 0"]
        """
        df["tags/sentence"] = df["tags"] / float(number_of_sentences)
        df["tags/sentence"] = df["tags/sentence"].apply(lambda x: "{:.2f}".format(x))

        # relative tags w/ 0
        number_of_occurrences = df["tags"].sum()
        df["tags relative w/ 0"] = df["tags"] / number_of_occurrences
        df["tags relative w/ 0"] = df["tags relative w/ 0"].apply(
            lambda x: "{:.2f}".format(x)
        )

        # relative tags w/o 0
        number_of_filtered_occurrences = df["tags"].sum() - df.loc["O"]["tags"]
        df["tags relative w/o 0"] = df["tags"] / number_of_filtered_occurrences
        df["tags relative w/o 0"]["O"] = 0
        df["tags relative w/o 0"] = df["tags relative w/o 0"].apply(
            lambda x: "{:.2f}".format(x)
        )

        df = df.reindex(
            ["tags", "tags/sentence", "tags relative w/ 0", "tags relative w/o 0"],
            axis=1,
        )

        return df

    @staticmethod
    def _get_num_tokens(df: pd.DataFrame) -> int:
        """
        Args:
            df: DataFrame with column "tags"

        Returns:
            num_tokens: sum of all values in column "tags"
        """
        assert (
            "tags" in df.columns
        ), f"ERROR! df.columns = {df.columns} does not contain 'tags'"
        return df.loc[:, "tags"].sum()

    def _write_log_file(
        self, num_sentences_total: int, num_tokens_total: int
    ) -> None:  # pragma: no cover
        # print/log
        log_file = join(self.dataset_path, "analyze_data", f"{self.ner_dataset}.log")
        default_logger = DefaultLogger(
            __file__, log_file=log_file, level="info", mode="w"
        )
        for phase in self.phases:
            default_logger.log_info("")
            default_logger.log_info(f">>> {phase} <<<<")
            default_logger.log_info(
                f"num_sentences = {self.num_sentences[phase]} "
                f"({100 * self.num_sentences[phase] / num_sentences_total:.2f}% of total = {num_sentences_total})"
            )
            default_logger.log_info(
                f"num_tokens = {self.num_tokens[phase]} "
                f"({100 * self.num_tokens[phase] / num_tokens_total:.2f}% of total = {num_tokens_total})"
            )
            default_logger.log_info(self.stats_aggregated[phase])

        default_logger.log_info("")
        default_logger.log_info(f"num_sentences = {self.num_sentences}")
        default_logger.log_info(f"num_tokens = {self.num_tokens}")
        default_logger.log_info(f">>> total <<<<")
        default_logger.log_info(self.stats_aggregated["total"])



import pytest
from typing import Union, Dict, Optional, List

from pkg_resources import resource_filename
import os

from nerblackbox.modules.experiment_config.experiment import Experiment
from nerblackbox.modules.utils.util_functions import get_run_name

os.environ["DATA_DIR"] = resource_filename("nerblackbox", f"tests/test_data")


class TestExperiment:

    experiment = Experiment(
        experiment_name="test_experiment",
        from_config=True,
        run_name="runA",
        device="gpu",
        fp16=True,
    )

    # 1 ################################################################################################################
    @pytest.mark.parametrize(
        "runs_name_nr, runs_params, runs_hparams",
        [
            (
                ["runA-1"],
                {
                    "runA-1": {
                        "experiment_name": "test_experiment",
                        "from_config": True,
                        "run_name": "runA",
                        "run_name_nr": "runA-1",
                        "device": "gpu",
                        "fp16": True,
                        "experiment_run_name_nr": "test_experiment/runA-1",
                        # params
                        "dataset_name": "swedish_ner_corpus",  # update non-default
                        "annotation_scheme": "plain",  # update non-default
                        "prune_ratio_train": 0.01,  # update non-default
                        "prune_ratio_val": 0.01,  # update non-default
                        "prune_ratio_test": 0.01,  # update non-default
                        "train_on_val": True,  # update non-default
                        "train_on_test": True,  # update non-default
                        "checkpoints": True,
                        "logging_level": "info",
                        "multiple_runs": 1,
                        "seed": 42,
                        "pretrained_model_name": "KB/bert-base-swedish-cased",  # update non-default
                        "uncased": False,  # update non-default
                    }
                },
                {
                    "runA-1": {
                        # hparams
                        "batch_size": 16,
                        "max_seq_length": 64,
                        "max_epochs": 3,  # update non-default
                        "early_stopping": False,  # update non-default
                        "monitor": "val_loss",
                        "mode": "min",
                        "min_delta": 0.0,
                        "patience": 0,
                        "lr_warmup_epochs": 2,
                        # "lr_cooldown_epochs": 3,
                        "lr_cooldown_restarts": False,
                        "lr_num_cycles": 4,
                        "lr_max": 2e-5,  # update non-default
                        "lr_schedule": "hybrid",  # update non-default
                        "lr_cooldown_epochs": 1,  # run-specific
                    }
                },
            )
        ],
    )
    def test_create_info_for_runs(
        self,
        runs_name_nr: List[str],
        runs_params: Dict[str, Dict[str, Union[str, int, float, bool]]],
        runs_hparams: Dict[str, Dict[str, Union[str, int, float, bool]]],
    ):
        # 1. runs_name_nr
        assert (
            self.experiment.runs_name_nr == runs_name_nr
        ), f"ERROR! self.experiment.runs_name_nr = {self.experiment.runs_name_nr} != {runs_name_nr}"

        # 2. runs_params
        assert sorted(list(self.experiment.runs_params.keys())) == sorted(
            list(runs_params.keys())
        ), (
            f"ERROR! self.experiment.runs_params.keys() = {sorted(list(self.experiment.runs_params.keys()))} "
            f"does not equal runs_params.keys() = {sorted(list(runs_params.keys()))}"
        )
        keys = list(self.experiment.runs_params.keys())
        for key in keys:
            assert self.experiment.runs_params[key] == runs_params[key], (
                f"ERROR! self.experiment.runs_params[{key}] = {self.experiment.runs_params[key]} "
                f"!= {self.experiment.runs_params[key]}"
            )

    # 2 ################################################################################################################
    @pytest.mark.parametrize(
        "run_name_nr, params_and_hparams",
        [
            (
                None,
                {
                    # params
                    "dataset_name": "swedish_ner_corpus",  # update non-default
                    "annotation_scheme": "plain",  # update non-default
                    "prune_ratio_train": 0.01,  # update non-default
                    "prune_ratio_val": 0.01,  # update non-default
                    "prune_ratio_test": 0.01,  # update non-default
                    "train_on_val": True,  # update non-default
                    "train_on_test": True,  # update non-default
                    "checkpoints": True,
                    "logging_level": "info",
                    "multiple_runs": 1,
                    "seed": 42,
                    "pretrained_model_name": "KB/bert-base-swedish-cased",  # update non-default
                    "uncased": False,  # update non-default
                    # hparams
                    "batch_size": 16,
                    "max_seq_length": 64,
                    "max_epochs": 3,  # update non-default
                    "early_stopping": False,  # update non-default
                    "monitor": "val_loss",
                    "mode": "min",
                    "min_delta": 0.0,
                    "patience": 0,
                    "lr_warmup_epochs": 2,
                    "lr_cooldown_epochs": 3,
                    "lr_cooldown_restarts": False,
                    "lr_num_cycles": 4,
                    "lr_max": 2e-5,  # update non-default
                    "lr_schedule": "hybrid",  # update non-default
                },
            ),
            (
                "runA",
                {
                    "lr_cooldown_epochs": 1,
                },
            ),
        ],
    )
    def test_create_info_for_mlflow(
        self,
        run_name_nr: Optional[str],
        params_and_hparams: Dict[str, Union[str, int, float, bool]],
    ):
        if run_name_nr is None:
            test_params_and_hparams = self.experiment.params_and_hparams["general"]
        else:
            test_params_and_hparams = self.experiment.params_and_hparams[
                get_run_name(run_name_nr)
            ]
        assert sorted(list(test_params_and_hparams.keys())) == sorted(
            list(params_and_hparams.keys())
        ), (
            f"ERROR! test_params_and_hparams.keys() = {sorted(list(test_params_and_hparams.keys()))} does not equal "
            f"params_and_hparams.keys() = {sorted(list(params_and_hparams.keys()))}"
        )

        keys = list(test_params_and_hparams.keys())
        for key in keys:
            assert test_params_and_hparams[key] == params_and_hparams[key], (
                f"ERROR! test_params_and_hparams[{key}] = {test_params_and_hparams[key]} "
                f"!= {params_and_hparams[key]}"
            )



from typing import List, Dict, Union, Any, Tuple
from nerblackbox.modules.ner_training.annotation_tags.tags import Tags
from copy import deepcopy


class TokenTags:
    def __init__(
        self, token_tag_list: List[Dict[str, str]], scheme: str, level: str = "token"
    ):
        """
        Args:
            token_tag_list: e.g. [
                {"char_start": "0", "char_end": "7", "token": "example", "tag": "I-TAG"},
                ..
            ]
            scheme: e.g. "plain" or "bio"
        """
        self.token_tag_list: List[Dict[str, str]] = token_tag_list
        self.scheme = scheme
        self.level = level

        if self.level in ["token", "word"]:
            self._assert_scheme_consistency()

    def _assert_scheme_consistency(self):
        """
        assert that tags in self.token_tag_list are in accordance with self.scheme
        """
        tags = [elem["tag"] for elem in self.token_tag_list if elem["tag"] != "O"]
        if len(tags) == 0:
            possible_schemes = ["plain", "bio", "bilou"]
        elif all(["-" not in elem for elem in tags]):
            possible_schemes = ["plain"]
        elif all(["-" in elem for elem in tags]):
            possible_schemes = ["bio", "bilou"]
        else:
            raise Exception(
                "ERROR! inconsistent tags found. they do not seem to belong to a well-defined scheme."
            )

        assert (
            self.scheme in possible_schemes
        ), f"ERROR! scheme = {self.scheme} is inconsistent with possible_schemes = {possible_schemes}!"

    def as_list(self):
        return self.token_tag_list

    ####################################################################################################################
    # MAIN METHODS
    ####################################################################################################################
    def restore_annotation_scheme_consistency(self) -> None:
        """
        plain tags: not modified
        bio   tags: restore annotation scheme consistency
        bilou tags: restore annotation scheme consistency

        Changed Attr:
            token_tag_list: List[Dict[str, str]], e.g.
            [
                {"char_start": "0", "char_end": "7", "token": "example", "tag": "I-TAG"},
                ..
            ]
            --->
            [
                {"char_start": "0", "char_end": "7", "token": "example", "tag": "B-TAG"},
                ..
            ]
        """
        if self.scheme == "plain":
            pass
        elif self.scheme in ["bio", "bilou"]:
            token_tag_list_restored: List[Dict[str, str]] = list()
            for i in range(len(self.token_tag_list)):
                example_word_prediction_restored = self.token_tag_list[i]
                current_tag = self.token_tag_list[i]["tag"]
                current_tag = self._assert_str(current_tag, "current_tag")
                previous_tag = self.token_tag_list[i - 1]["tag"] if i > 0 else None

                if self.scheme == "bio":
                    example_word_prediction_restored["tag"] = Tags.convert_tag_bio2bio(
                        current_tag, previous=previous_tag
                    )[0]
                else:
                    subsequent_tag = (
                        self.token_tag_list[i + 1]["tag"]
                        if i < len(self.token_tag_list) - 1
                        else None
                    )
                    example_word_prediction_restored[
                        "tag"
                    ] = Tags.convert_tag_bilou2bilou(
                        current_tag, previous=previous_tag, subsequent=subsequent_tag
                    )[
                        0
                    ]

                token_tag_list_restored.append(example_word_prediction_restored)

            assert len(token_tag_list_restored) == len(self.token_tag_list), f"ERROR!"

            self.token_tag_list = token_tag_list_restored
        else:
            raise Exception(
                f"ERROR! restore annotation scheme consistency "
                f"not implemented for scheme = {self.scheme}."
            )

    def merge_tokens_to_words(self) -> None:
        """
        discard tokens that are not first token of a word

        Changed Attr:
            token_tag_list: List[Dict[str, str]], e.g.
            [
                {"char_start": "0", "char_end": "4", "token": "2020", "tag": "B-TAG"},
                {"char_start": "4", "char_end": "5", "token": "-, "tag": "I-TAG"},
                {"char_start": "5", "char_end": "7", "token": "04, "tag": "I-TAG"},
                ..
            ]
            --->
            [
                {"char_start": "0", "char_end": "7", "token": "2020-04", "tag": "B-TAG"},
                ..
            ]
            level: 'word'

        """
        for i in range(len(self.token_tag_list) - 1, 0, -1):
            if (
                i > 0
                and self.token_tag_list[i]["char_start"]
                == self.token_tag_list[i - 1]["char_end"]
            ):
                self.token_tag_list[i]["tag"] = "DELETE"
                self.token_tag_list[i - 1]["char_end"] = self.token_tag_list[i][
                    "char_end"
                ]
                self.token_tag_list[i - 1]["token"] += self.token_tag_list[i]["token"]

        self.token_tag_list = [
            elem for elem in self.token_tag_list if elem["tag"] != "DELETE"
        ]
        self.level = "word"

    def unpretokenize(self, _pretokenization_offsets: List[Tuple[int, int]]):
        """
        revert pretokenization using pretokenization offsets.

        Args:
            _pretokenization_offsets: e.g. [(0,4), (4,5), (5,7), (7,8), (8,10), (11,15)]

        Changed Attr:
            token_tag_list: List[Dict[str, str]], e.g.
                [
                    {'char_start': '0', 'char_end': '4', 'token': '2021', 'tag': 'B-PI'},
                    {'char_start': '5', 'char_end': '6', 'token': '-', 'tag': 'I-PI'},
                    {'char_start': '7', 'char_end': '9', 'token': '10', 'tag': 'I-PI'},
                    {'char_start': '10', 'char_end': '11', 'token': '-', 'tag': 'I-PI'},
                    {'char_start': '12', 'char_end': '14', 'token': '14', 'tag': 'I-PI'},
                    {'char_start': '15', 'char_end': '20', 'token': 'Mamma', 'tag': 'O'},
                ]
            --->
                [
                    {'char_start': '0', 'char_end': '4', 'token': '2021', 'tag': 'B-PI'},
                    {'char_start': '4', 'char_end': '5', 'token': '-', 'tag': 'I-PI'},
                    {'char_start': '5', 'char_end': '7', 'token': '10', 'tag': 'I-PI'},
                    {'char_start': '7', 'char_end': '8', 'token': '-', 'tag': 'I-PI'},
                    {'char_start': '8', 'char_end': '10', 'token': '14', 'tag': 'I-PI'},
                    {'char_start': '11', 'char_end': '15', 'token': 'Mamma', 'tag': 'O'},
                ]
        """
        assert len(self.token_tag_list) == len(_pretokenization_offsets), (
            f"ERROR! #token_tag_list = {len(self.token_tag_list)} != "
            f"#pretokenization_offsets = {len(_pretokenization_offsets)}"
        )

        nr_tokens = len(self.token_tag_list)
        for j in range(nr_tokens):
            self.token_tag_list[j]["char_start"] = str(_pretokenization_offsets[j][0])
            self.token_tag_list[j]["char_end"] = str(_pretokenization_offsets[j][1])

    def merge_tokens_to_entities(self, original_text: str, verbose: bool) -> None:
        """
        plain tags:
            - discard tokens with tag 'O'
        bio   tags:
            - merge token predictions that belong together (B-* & I-*)
            - discard tokens with tag 'O'
        bilou tags:
            - merge token predictions that belong together (B-* & I-* & L-*; U-*)
            - discard tokens with tag 'O'

        Args:
            original_text: e.g. 'example sentence.'
            verbose: e.g. False

        Changed Attr:
            token_tag_list: List[Dict[str, str]], e.g.
            [
                {"char_start": "0", "char_end": "7", "token": "example", "tag": "B-TAG"},
                {"char_start": "8", "char_end": "16", "token": "sentence", "tag": "I-TAG"},
                {"char_start": "17", "char_end": "18", "token": ".", "tag": "O"},
                ..
            ]
            --->
            [
                {"char_start": "0", "char_end": "16", "token": "example sentence", "tag": "TAG"},
                ..
            ]
            level: 'entity'
        """
        count = {
            "o_tags": 0,
            "replace": 0,
            "drop": 0,
            "merge": 0,
            "unmodified": 0,
        }
        entity_prefixes_without_b = {
            "bio": ["I-"],
            "bilou": ["I-", "L-"],
        }

        threshold = 0
        merged_ner_tags = list()
        for i in range(len(self.token_tag_list)):
            current_tag = self.token_tag_list[i]["tag"]
            current_tag = self._assert_str(current_tag, "current_tag")
            n_tags = 0
            if current_tag == "O":
                count["o_tags"] += 1
            else:
                merged_ner_tag = None
                if i >= threshold:
                    if self.scheme == "plain":
                        for n in range(i + 1, len(self.token_tag_list)):
                            subsequent_tag = self.token_tag_list[n]["tag"]
                            subsequent_tag = self._assert_str(
                                subsequent_tag, "subsequent_tag"
                            )
                            if subsequent_tag == current_tag:
                                n_tags += 1
                            else:
                                threshold = n
                                break
                            if n == len(self.token_tag_list) - 1:
                                threshold = n + 1  # such that last element is dropped
                        merged_ner_tag = self._merge_tokens(i, original_text, n_tags)
                    elif self.scheme in ["bio", "bilou"]:
                        if current_tag.startswith("B-"):  # BIO scheme
                            plain = current_tag.split("-")[-1]
                            for n in range(i + 1, len(self.token_tag_list)):
                                subsequent_tag = self.token_tag_list[n]["tag"]
                                subsequent_tag = self._assert_str(
                                    subsequent_tag, "subsequent_tag"
                                )
                                if (
                                    len(subsequent_tag) > 2
                                    and subsequent_tag[:2]
                                    in entity_prefixes_without_b[self.scheme]
                                    and subsequent_tag[2:] == plain
                                ):
                                    n_tags += 1
                                    if subsequent_tag[:2] in [
                                        "L-"
                                    ]:  # only applies to "bilou"
                                        # end of entity
                                        threshold = n + 1
                                        break
                                else:
                                    # end of entity
                                    threshold = n
                                    break
                                if n == len(self.token_tag_list) - 1:
                                    threshold = (
                                        n + 1
                                    )  # such that last element is dropped
                            merged_ner_tag = self._merge_tokens(
                                i, original_text, n_tags
                            )
                        elif current_tag.startswith("I-"):
                            count["drop"] += 1
                        elif current_tag.startswith("L-"):  # only applies to "bilou"
                            count["drop"] += 1
                        elif current_tag.startswith("U-"):  # only applies to "bilou"
                            merged_ner_tag = self._merge_tokens(
                                i, original_text, n_tags
                            )

                if merged_ner_tag is not None:
                    merged_ner_tags.append(merged_ner_tag)
                    if n_tags == 0:
                        if self.scheme == "plain":
                            count["unmodified"] += 1
                        else:
                            count["replace"] += 1
                    else:
                        count["merge"] += 1 + n_tags

        assert count["o_tags"] + count["replace"] + count["drop"] + count[
            "merge"
        ] + count["unmodified"] == len(
            self.token_tag_list
        ), f"{count} -> {sum(count.values())} != {len(self.token_tag_list)} | {self.token_tag_list}"

        """
        if count["merge"] > 0 or count["replace"] > 0:
            assert (
                    count["unmodified"] == 0
            ), f"{count} -> if merge or replaced are > 0, unmodified should be == 0."

        if count["unmodified"] > 0:
            assert (
                    count["merge"] == 0
            ), f"{count} -> if unmodified is > 0, merge should be == 0."
            assert (
                    count["replace"] == 0
            ), f"{count} -> if unmodified is > 0, replace should be == 0."
        """

        self.token_tag_list = merged_ner_tags
        if verbose:
            print(
                f"> merged {len(self.token_tag_list)} BIO-tags "
                f"(simple replace: {count['replace']}, "
                f"merge: {count['merge']}, "
                f"O-tags: {count['o_tags']}, "
                f"drop: {count['drop']}, "
                f"unmodified: {count['unmodified']}).\n"
            )

        self.level = "entity"

    ####################################################################################################################
    # HELPER
    ####################################################################################################################
    @staticmethod
    def _assert_str(_object: Union[str, Dict[Any, Any]], _object_name: str) -> str:
        assert isinstance(
            _object, str
        ), f"ERROR! {_object_name} = {_object} is {type(_object)} but should be string"
        return _object

    def _merge_tokens(
        self, _index: int, _original_text: str, _n_tags: int
    ) -> Dict[str, str]:
        """
        merge tokens [_index: _index+_ntags] using self.token_tag_list

        Args:
            _index:         e.g. 1
            _original_text: e.g. 'annotera den här texten'
            _n_tags:        e.g. 1

        Return:
            merged_token: e.g. {'char_start': '9', 'char_end': '16', 'token': 'den här', 'tag': 'ORG'}
        """
        merged_ner_tag = deepcopy(self.token_tag_list[_index])
        assert isinstance(
            merged_ner_tag["tag"], str
        ), "ERROR! merged_ner_tag.tag should be a string"

        # 1. replace tag
        merged_ner_tag["tag"] = merged_ner_tag["tag"].split("-")[-1]

        if _n_tags > 0:
            # 2. replace char_end
            merged_ner_tag["char_end"] = self.token_tag_list[_index + _n_tags][
                "char_end"
            ]

            assert isinstance(
                merged_ner_tag["char_start"], str
            ), "ERROR! merged_ner_tag.char_start should be a string"
            assert isinstance(
                merged_ner_tag["char_end"], str
            ), "ERROR! merged_ner_tag.char_end should be a string"

            # 3. replace token
            merged_ner_tag["token"] = _original_text[
                int(merged_ner_tag["char_start"]) : int(merged_ner_tag["char_end"])
            ]
        return merged_ner_tag



import numpy as np
from typing import List, Optional, Tuple


class Tags:
    def __init__(self, tag_list: List[str]):
        """
        :param tag_list: e.g. ["O", "PER", "O", "ORG"] or ["O", "B-person", "I-person", "O"]
        """
        self.tag_list = tag_list

    ####################################################################################################################
    # 1. MAIN METHODS
    ####################################################################################################################
    def convert_scheme(self, source_scheme: str, target_scheme: str) -> List[str]:
        """
        Args:
            source_scheme: e.g. 'plain', 'bio'
            target_scheme: e.g. 'bio', 'plain'

        Uses:
            tag_list:           e.g. ['O',   'ORG',   'ORG'] or ['O', 'B-ORG', 'I-ORG']

        Returns:
            tag_list_converted: e.g. ['O', 'B-ORG', 'I-ORG'] or ['O',   'ORG',   'ORG']
        """
        if source_scheme == "plain":
            self._assert_plain_tags()
        elif source_scheme == "bio":
            self._assert_bio_tags()
        elif source_scheme == "bilou":
            self._assert_bilou_tags()
        else:
            raise Exception(f"ERROR! source scheme = {source_scheme} not implemented.")

        if source_scheme == target_scheme:  # "plain" and target_scheme == "plain":
            return list(self.tag_list)
        elif source_scheme == "plain" and target_scheme == "bio":
            return self._convert_tags_plain2bio()
        elif source_scheme == "plain" and target_scheme == "bilou":
            return self._convert_tags_plain2bilou()
        elif source_scheme in ["bio", "bilou"] and target_scheme == "plain":
            return self._convert_tags_bio2plain()
        elif source_scheme == "bio" and target_scheme == "bilou":
            return self.restore_annotation_scheme_consistency(scheme=target_scheme)[
                0
            ]  # conversion
        elif source_scheme == "bilou" and target_scheme == "bio":
            return self._convert_tags_bilou2bio()
        else:
            raise Exception(
                f"ERROR! source & target scheme = {source_scheme} & {target_scheme} not implemented."
            )

    def get_annotation_scheme_abidance(
        self, scheme: str
    ) -> float:  # Tuple[int, int, float]:
        """
        compute how many tokens in self.tag_list abide to the annotation scheme

        Returns:
            annotation_scheme_abidance:
                - absolute number of correct tokens
                - total number of tokens
                - relative number of correct tokens
        """
        # return 49.0, 50.0, 0.98
        return 0.98

    ####################################################################################################################
    # 2. HELPER METHODS
    ####################################################################################################################
    def _assert_plain_tags(self) -> None:
        for i, tag in enumerate(self.tag_list):
            if tag != "O" and (len(tag) > 2 and tag[1] == "-"):
                raise Exception(
                    f"ERROR! tags do not have expected plain format. tag = {tag} encountered at index {i}."
                )

    def _assert_bio_tags(self) -> None:
        for i, tag in enumerate(self.tag_list):
            if tag != "O" and (
                len(tag) <= 2 or tag[0] not in ["B", "I"] or tag[1] != "-"
            ):
                raise Exception(
                    f"ERROR! tags do not have expected bio format. tag = {tag} encountered at index {i}."
                )

    def _assert_bilou_tags(self) -> None:
        for i, tag in enumerate(self.tag_list):
            if tag != "O" and (
                len(tag) <= 2 or tag[0] not in ["B", "I", "L", "U"] or tag[1] != "-"
            ):
                raise Exception(
                    f"ERROR! tags do not have expected bilou format. tag = {tag} encountered at index {i}."
                )

    def _convert_tags_plain2bio(self) -> List[str]:
        """
        adds bio prefixes to plain tags

        Uses:
            tag_list:     e.g. ['O',   'ORG',   'ORG']

        Returns:
            bio_tag_list: e.g. ['O', 'B-ORG', 'I-ORG']
        """
        return [
            self._convert_tag_plain2bio(
                self.tag_list[i], previous=self.tag_list[i - 1] if i > 0 else None
            )
            for i in range(len(self.tag_list))
        ]

    @staticmethod
    def _convert_tag_plain2bio(tag: str, previous: Optional[str] = None) -> str:
        """
        add bio prefix to plain tag, depending on previous tag

        Args:
            tag:      e.g. 'ORG'
            previous: e.g. 'ORG'

        Returns:
            bio_tag:  e.g. 'I-ORG'
        """
        if tag == "O":
            return tag
        elif previous is None or tag != previous:
            return f"B-{tag}"
        else:
            return f"I-{tag}"

    def _convert_tags_plain2bilou(self) -> List[str]:
        """
        adds bilou prefixes to plain tags

        Uses:
            tag_list:     e.g. ['O',   'ORG',   'ORG']

        Returns:
            bio_tag_list: e.g. ['O', 'B-ORG', 'I-ORG']
        """
        return [
            self._convert_tag_plain2bilou(
                self.tag_list[i],
                previous=self.tag_list[i - 1] if i > 0 else None,
                subsequent=self.tag_list[i + 1] if i < len(self.tag_list) - 1 else None,
            )
            for i in range(len(self.tag_list))
        ]

    @staticmethod
    def _convert_tag_plain2bilou(
        tag: str, previous: Optional[str] = None, subsequent: Optional[str] = None
    ) -> str:
        """
        add bilou prefix to plain tag, depending on previous and subsequent tag

        Args:
            tag:        e.g. 'ORG'
            previous:   e.g. 'ORG'
            subsequent: e.g. 'ORG'

        Returns:
            bilou_tag:  e.g. 'I-ORG'
        """
        if tag == "O":
            return tag
        else:
            previous_condition = previous is None or tag != previous
            subsequent_condition = subsequent is None or tag != subsequent
            if previous_condition and subsequent_condition:
                return f"U-{tag}"
            elif previous_condition and not subsequent_condition:
                return f"B-{tag}"
            elif not previous_condition and subsequent_condition:
                return f"L-{tag}"
            else:
                return f"I-{tag}"

    def _convert_tags_bio2plain(self) -> List[str]:
        """
        retrieve plain tags by removing bio prefixes

        Uses:
            bio_tag_list: e.g. ['O', 'B-ORG', 'I-ORG']

        Returns:
            tag_list:     e.g. ['O',   'ORG',   'ORG']
        """
        return [elem.split("-")[-1] for elem in self.tag_list]

    def _convert_tags_bilou2bio(self) -> List[str]:
        """
        convert bilou to bio tags

        Uses:
            bilou_tag_list: e.g. ['O', 'B-ORG', 'L-ORG']

        Returns:
            bio_tag_list:   e.g. ['O', 'B-ORG', 'I-ORG']
        """
        return [self._convert_tag_bilou2bio(elem) for elem in self.tag_list]

    @staticmethod
    def _convert_tag_bilou2bio(tag: str):
        """
        Args:
            tag:      e.g. 'L-ORG'

        Returns:
            bio_tag:  e.g. 'I-ORG'
        """
        if tag == "O" or tag.startswith("B-") or tag.startswith("I-"):
            return tag
        else:
            plain = tag.split("-")[-1]
            if tag.startswith("U-"):
                return f"B-{plain}"
            elif tag.startswith("L-"):
                return f"I-{plain}"

    def restore_annotation_scheme_consistency(
        self, scheme: str
    ) -> Tuple[List[str], float]:
        """

        Args:
            scheme:

        Uses:
            tag_list:           e.g. ['O',   'ORG',   'ORG'] or ['O', 'B-ORG', 'I-ORG']

        Returns:
            tag_list_converted: e.g. ['O', 'B-ORG', 'I-ORG'] or ['O',   'ORG',   'ORG']
            asr_abidance:       e.g. 0.98
                = how many tokens in self.tag_list abide to the annotation scheme
                = absolute number of correct tokens / total number of tokens
        """
        if scheme == "bio":
            converted_tuples = [
                self.convert_tag_bio2bio(
                    self.tag_list[i], previous=self.tag_list[i - 1] if i > 0 else None
                )
                for i in range(len(self.tag_list))
            ]
        elif scheme == "bilou":
            converted_tuples = [
                self.convert_tag_bilou2bilou(
                    self.tag_list[i],
                    previous=self.tag_list[i - 1] if i > 0 else None,
                    subsequent=self.tag_list[i + 1]
                    if i < len(self.tag_list) - 1
                    else None,
                )
                for i in range(len(self.tag_list))
            ]
        else:
            raise Exception(
                f"ERROR! restore annotation scheme consistency not implemented for scheme = {scheme}."
            )

        tag_list_converted = [tup[0] for tup in converted_tuples]
        asr_abidance = float(1 - np.average([tup[1] for tup in converted_tuples]))
        return tag_list_converted, asr_abidance

    ####################################################################################################################
    # 3. CLASS METHODS
    ####################################################################################################################
    @classmethod
    def convert_tag_bio2bio(
        cls, current: str, previous: Optional[str] = None
    ) -> Tuple[str, bool]:
        """
        correct bio prefix, depending on previous tag

        Args:
            current:  e.g. 'I-ORG'
            previous: e.g. 'O'

        Returns:
            current_corrected:  e.g. 'B-ORG'
            bool_corrected:     e.g. True
        """
        if current == "O" or current.startswith("B-"):
            return current, False
        else:
            assert (
                current[0] in ["I"] and current[1] == "-"
            ), f"ERROR! bio tag {current} should be of the format O, B-*, I-*"

            plain = current.split("-")[-1]
            condition_previous = previous is None or previous not in [
                f"B-{plain}",
                f"I-{plain}",
            ]

            if condition_previous:
                # rule 1: !I/!B + I -> B
                return f"B-{plain}", True
            else:
                return f"I-{plain}", False

    @classmethod
    def convert_tag_bilou2bilou(
        cls,
        current: str,
        previous: Optional[str] = None,
        subsequent: Optional[str] = None,
    ) -> Tuple[str, bool]:
        """
        correct bilou prefix, depending on previous and next tag

        Args:
            current:    e.g. 'I-ORG'
            previous:   e.g. 'O'
            subsequent: e.g. 'O'

        Returns:
            current_corrected:  e.g. 'U-ORG'
            bool_corrected:     e.g. True
        """
        if current == "O" or current.startswith("U-"):
            return current, False
        else:
            assert (
                current[0] in ["B", "I", "L"] and current[1] == "-"
            ), f"ERROR! bilou tag {current} should be of the format O, B-*, I-*, L-*, U-*"

            plain = current.split("-")[-1]
            condition_previous = previous is None or previous not in [
                f"B-{plain}",
                f"I-{plain}",
            ]
            condition_subsequent = subsequent is None or subsequent not in [
                f"I-{plain}",
                f"L-{plain}",
            ]
            bool_corrected = False
            while 1:
                if current.startswith("I-"):
                    if condition_subsequent:
                        # rule 1: I + !I/!L -> L
                        current = f"L-{plain}"
                        bool_corrected = True
                    elif condition_previous:
                        # rule 2: !B/!I + I -> B
                        current = f"B-{plain}"
                        bool_corrected = True
                    else:
                        break
                elif current.startswith("L-"):
                    if condition_previous:
                        # rule 3: !B/!I + L -> B
                        current = f"B-{plain}"
                        bool_corrected = True
                    else:
                        break
                elif current.startswith("B-"):
                    if condition_subsequent:
                        # rule 4: B + !I/!L -> U
                        current = f"U-{plain}"
                        bool_corrected = True
                    else:
                        break
                elif current.startswith("U-"):
                    break
                else:
                    raise Exception(
                        f"ERROR! this should not happen. current = {current}"
                    )

            return current, bool_corrected



from typing import Optional, Dict, Any


class Utils:
    @classmethod
    def process_kwargs_optional(
        cls, _kwargs_optional: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        general helper function
        filters out key-value pairs that have value = None

        Args:
            _kwargs_optional: e.g. {"a": 1, "b": None}

        Returns:
            _kwargs:          e.g. {"a": 1}
        """
        if _kwargs_optional is None:
            return {}
        else:
            return {k: v for k, v in _kwargs_optional.items() if v is not None}

    @classmethod
    def extract_hparams(cls, _kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Args:
            _kwargs: e.g. {"a": 1, "run_name": "runA-1"}

        Returns:
            _hparams: e.g. {"a": 1}
        """
        # hparams
        exclude_keys = ["experiment_name", "run_name", "device", "fp16"]
        _hparams = {
            _key: _kwargs[_key] for _key in _kwargs.keys() if _key not in exclude_keys
        }

        return _hparams
